from typing import Optional, Dict, Union

import torch
from torch.utils.data import DataLoader

from torch_mist.estimators.base import MIEstimator
from torch_mist.utils.evaluation.model import evaluate

from torch_mist.utils.data.utils import (
    TensorDictLike,
    update_dataloader,
    make_default_dataloader,
)


def evaluate_mi(
    estimator: MIEstimator,
    data: TensorDictLike,
    device: torch.device = torch.device("cpu"),
    batch_size: Optional[int] = None,
    num_workers: int = 0,
) -> Union[float, Dict[str, float]]:
    # Make a dataloader if necessary
    dataloader = make_default_dataloader(
        data=data,
        batch_size=batch_size,
        num_workers=num_workers,
    )

    dataloader = update_dataloader(estimator, dataloader)

    values = evaluate(
        model=estimator,
        method="mutual_information",
        data=dataloader,
        device=device,
    )

    if isinstance(values, dict):
        values = {
            f"I({x_key};{y_key})": v for (x_key, y_key), v in values.items()
        }

    return values



from abc import abstractmethod
from typing import Dict, Union, Callable, Tuple, List

import torch
import torch.nn as nn

from torch_mist.nn import Model


class MIEstimator(Model):
    infomax_gradient: Dict[str, bool] = {"x": False, "y": False}

    def __init__(self):
        super().__init__()
        self._components_to_pretrain: List[Tuple[Callable, nn.Module]] = []

    @abstractmethod
    def log_ratio(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError()

    @abstractmethod
    def unnormalized_log_ratio(
        self, x: torch.Tensor, y: torch.Tensor
    ) -> torch.Tensor:
        raise NotImplementedError()

    def mutual_information(
        self, x: torch.Tensor, y: torch.Tensor
    ) -> Union[torch.Tensor, Dict[str, torch.Tensor]]:
        log_ratio = self.log_ratio(x, y)
        assert (
            not isinstance(x, torch.LongTensor)
            and (log_ratio.shape == x.shape[:-1])
        ) or (isinstance(x, torch.LongTensor) and log_ratio.shape == x.shape)
        return log_ratio.mean()

    @abstractmethod
    def batch_loss(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError()

    def loss(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        batch_loss = self.batch_loss(x=x, y=y)
        assert (
            (not isinstance(x, torch.LongTensor))
            and (batch_loss.shape == x.shape[:-1])
        ) or (isinstance(x, torch.LongTensor) and batch_loss.shape == x.shape)
        return batch_loss.mean()

    def forward(
        self, x: torch.Tensor, y: torch.Tensor
    ) -> Union[torch.Tensor, Dict[str, torch.Tensor]]:
        return self.loss(x, y)



from typing import Optional

import torch

from torch_mist.baseline import BatchLogMeanExp
from torch_mist.critic import Critic
from torch_mist.estimators.discriminative.base import (
    BaselineDiscriminativeMIEstimator,
)


class InfoNCE(BaselineDiscriminativeMIEstimator):
    def __init__(self, critic: Critic, neg_samples: int = 0):
        # Note that this can be equivalently obtained by extending TUBA with a BatchLogMeanExp(dim=1) baseline
        # This implementation saves some computation
        super().__init__(
            critic=critic,
            neg_samples=neg_samples,  # 0 signifies the whole batch is used as negative samples
            baseline=BatchLogMeanExp("first"),
        )

    def _approx_log_partition(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        f_: torch.tensor,
        log_w: Optional[torch.Tensor],
    ):
        # Add the log_weights if provided
        if not (log_w is None):
            assert log_w.ndim == f_.ndim
            f_ = f_ + log_w

        # We override the compute_log_normalization just for efficiency
        # The result would be the same as the TUBA implementation with BatchLogMeanExp('first') baseline
        # We override the compute_log_normalization for efficiency since e^(F(x,y))-b(x) = 1
        log_Z = self.baseline(x=x, f_=f_)
        return log_Z



from .alpha_tuba import AlphaTUBA
from .flo import FLO
from .infonce import InfoNCE
from .js import JS
from .mine import MINE
from .nwj import NWJ
from .smile import SMILE
from .tuba import TUBA
from .dummy import DummyDiscriminativeMIEstimator



from typing import Dict, Any, Union, Optional

import numpy as np
import torch
from torch import nn

from torch_mist.decomposition.base import DimensionalityReduction
from torch_mist.estimators import (
    instantiate_estimator,
    TransformedMIEstimator,
)
from torch_mist.nn import dense_nn, Identity
from torch_mist.utils import train_mi_estimator
from torch_mist.utils.data.utils import TensorDictLike

DEFAULT_MAX_ITERATIONS = 5000
DEFAULT_BATCH_SIZE = 64


class MID(DimensionalityReduction):
    def _add_default_model_params(
        self, model_params: Optional[Dict[str, Any]]
    ) -> Dict[str, Any]:
        if model_params is None:
            model_params = {}
        if not ("hidden_dims" in model_params):
            model_params["hidden_dims"] = [128, 64]

        if not ("estimator_name" in model_params):
            model_params["estimator_name"] = "smile"
            if not ("neg_samples" in model_params):
                model_params["neg_samples"] = 8

        return model_params

    def _add_default_proj_params(
        self, proj_params: Optional[Dict[str, Any]]
    ) -> Dict[str, Any]:
        if proj_params is None:
            proj_params = {}
        if not ("hidden_dims" in proj_params):
            proj_params["hidden_dims"] = [128]
        if not ("nonlinearity" in proj_params):
            proj_params["nonlinearity"] = nn.ReLU(True)

        return proj_params

    def _instantiate_proj(self, x_dim: int) -> nn.Module:
        self.proj_params["input_dim"] = x_dim
        self.proj_params["output_dim"] = self.n_dim

        return dense_nn(**self.proj_params)

    def _instantiate_y_proj(self, y_dim: int) -> Optional[nn.Module]:
        if not (self.y_proj_params is None):
            self.y_proj_params["input_dim"] = y_dim
            return dense_nn(**self.y_proj_params)
        else:
            return None

    def _train_model(self, data: TensorDictLike, **train_params):
        return train_mi_estimator(self.model, data, **train_params)

    def _instantiate_model(self, x_dim: int, y_dim: int):
        self.model_params["x_dim"] = self.n_dim
        self.model_params["y_dim"] = y_dim
        mi_estimator = instantiate_estimator(**self.model_params)

        transforms = {"x": self.proj}
        if not (self.y_proj is None):
            transforms["y"] = self.y_proj

        return TransformedMIEstimator(
            transforms=transforms, base_estimator=mi_estimator
        )



import os
from typing import Tuple, List, Callable, Optional

import numpy as np
import torch
import hydra
from hydra.utils import instantiate
from omegaconf import OmegaConf, DictConfig
import random

from torch_mist.estimators import MIEstimator
from torch_mist.estimators.discriminative import DiscriminativeMIEstimator
from torch_mist.estimators.hybrid import HybridMIEstimator
from torch_mist.utils.data.utils import infer_dims, TensorDictLike
from torch_mist.utils.estimation import _instantiate_estimator, estimate_mi
from torch_mist.utils.logging.metrics import compute_mean_std


def default_logged_methods(
    estimator: MIEstimator,
) -> Tuple[List[Tuple[str, Callable]], List[Tuple[str, Callable]]]:
    eval_logged_methods = [
        ("log_ratio", compute_mean_std),
        ("batch_loss", compute_mean_std),
    ]

    if isinstance(estimator, DiscriminativeMIEstimator):
        eval_logged_methods += [
            ("approx_log_partition", compute_mean_std),
            ("unnormalized_log_ratio", compute_mean_std),
        ]

    if isinstance(estimator, HybridMIEstimator):
        eval_logged_methods += [
            ("generative_estimator.log_ratio", compute_mean_std),
            ("generative_estimator.batch_loss", compute_mean_std),
            ("discriminative_estimator.batch_loss", compute_mean_std),
            ("discriminative_estimator.log_ratio", compute_mean_std),
            (
                "discriminative_estimator.approx_log_partition",
                compute_mean_std,
            ),
            (
                "discriminative_estimator.unnormalized_log_ratio",
                compute_mean_std,
            ),
        ]

    train_logged_methods = [("batch_loss", compute_mean_std)]

    return train_logged_methods, eval_logged_methods


def prepare_data(
    conf: DictConfig,
) -> Tuple[
    TensorDictLike,
    Optional[TensorDictLike],
    Optional[TensorDictLike],
    Optional[float],
]:
    # Train (mandatory)
    data = instantiate(conf.data, _convert_="all")

    # Validation (optional)
    if "valid_data" in conf:
        valid_data = instantiate(conf.valid_data, _convert_="all")
    else:
        valid_data = None

    # Test (optional)
    if "test_data" in conf:
        test_data = instantiate(conf.test_data, _convert_="all")
    else:
        test_data = None

    # If the data is sampled from a distribution, compute the true mutual informaiton
    if "distribution" in conf:
        dist = instantiate(conf.distribution, _convert_="all")
        true_mi = dist.mutual_information()
    else:
        true_mi = None

    return data, valid_data, test_data, true_mi


@hydra.main(
    config_path="config", config_name="config.yaml", version_base="1.1"
)
def parse(conf: DictConfig):
    # Set the random seed
    if "seed" in conf:
        torch.manual_seed(conf.seed)
        np.random.seed(conf.seed)
        random.seed(conf.seed)

        if hasattr(conf.device, "tensor_cores"):
            torch.set_float32_matmul_precision(conf.device.matmul_precision)

    # Change the location to the original working directory (to fix issues with relative paths)
    hydra_wd = os.getcwd()
    try:
        os.chdir(hydra.utils.get_original_cwd())
    except Exception as e:
        print(e)
        pass

    # Instantiate the parameters and metadata if any
    if "params" in conf:
        conf.params = instantiate(conf.params)

    # Instantiate the loggers
    print("Instantiating the logger")
    logger = instantiate(conf.logger)

    # Add the configuration to the run if the logger allows
    if hasattr(logger, "add_config"):
        logger.add_config(OmegaConf.to_container(conf, resolve=True))

    # Instantiating the data
    print("Instantiating the Data")
    data, valid_data, test_data, true_mi = prepare_data(conf)

    # Infer the dimensionality of x and y if not specified by the config and update the values
    dims = infer_dims(data)
    conf.x_dim = dims[conf.estimation.x_key]
    conf.y_dim = dims[conf.estimation.y_key]

    # Instantiate the estimator
    print("Instantiating the Mutual Information Estimator")
    estimator = instantiate(conf.mi_estimator, _convert_="all")

    # Count and visualize the number of parameters
    n_parameters = 0
    for param in estimator.parameters():
        n_parameters += param.numel()
    print(f"{n_parameters} Parameters")

    # Define which quantities are logged
    train_logged_methods, eval_logged_methods = default_logged_methods(
        estimator
    )

    # Move back to the output directory created by hydra
    os.chdir(hydra_wd)

    # Train the estimator and evaluate the mutual information
    estimation_params = instantiate(conf.estimation, _convert_="all")
    estimated_mi, estimator, train_log = estimate_mi(
        estimator=estimator,
        data=data,
        valid_data=valid_data,
        test_data=test_data,
        logger=logger,
        train_logged_methods=train_logged_methods,
        eval_logged_methods=eval_logged_methods,
        return_estimator=True,
        **estimation_params,
    )

    if not (true_mi is None):
        print(f"True mi: {true_mi}")
    print(f"Estimated mi: {estimated_mi}")

    if conf.save_train_log:
        logger.save_log()

    if not (conf.save_trained_model is None):
        print(f"Saving the estimator in {conf.save_trained_model}")
        logger.save_model(estimator, conf.save_trained_model)


def run():
    OmegaConf.register_new_resolver("eval", eval)
    parse()


if __name__ == "__main__":
    run()



from typing import Optional, Dict

import torch
from torch.distributions import Distribution
from pyro.distributions import ConditionalDistribution

from torch_mist.estimators.generative.base import (
    ConditionalGenerativeMIEstimator,
)
from torch_mist.utils.caching import cached_method
from torch_mist.utils.freeze import is_trainable


class DoE(ConditionalGenerativeMIEstimator):
    infomax_gradient: Dict[str, bool] = {"x": True, "y": False}

    def __init__(
        self,
        q_Y_given_X: ConditionalDistribution,
        q_Y: Distribution,
    ):
        super().__init__(
            q_Y_given_X=q_Y_given_X,
        )
        self.q_Y = q_Y

    @cached_method
    def approx_log_p_y(
        self, y: torch.Tensor, x: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        log_q_y = self.q_Y.log_prob(y)
        assert (
            log_q_y.shape == y.shape[:-1] and torch.is_floating_point(y)
        ) or (log_q_y.shape == y.shape and not torch.is_floating_point(y))
        return log_q_y

    def batch_loss(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
    ) -> torch.Tensor:
        batch_loss = super().batch_loss(x, y)
        if is_trainable(self.q_Y):
            log_q_y = self.approx_log_p_y(x=x, y=y)
            batch_loss = batch_loss - log_q_y

            assert (
                batch_loss.shape == y.shape[:-1] and torch.is_floating_point(y)
            ) or (
                batch_loss.shape == y.shape and not torch.is_floating_point(y)
            )

        return batch_loss

    def __repr__(self):
        s = self.__class__.__name__ + "(\n"
        s += (
            "  "
            + "(q_Y_given_X): "
            + str(self.q_Y_given_X).replace("\n", "  \n")
            + "\n"
        )
        s += "  " + "(q_Y): " + str(self.q_Y).replace("\n", "  \n") + "\n"
        s += ")" + "\n"
        return s



from typing import Tuple

import torch


def expand_to_same_shape(*tensors) -> Tuple[torch.Tensor, ...]:
    if len(tensors) == 1:
        assert isinstance(tensors[0], torch.Tensor)
        return tensors[0]
    ndim = tensors[0].ndim

    # Tensors of integer are considered to have an extra (empty) dimension
    if torch.is_floating_point(tensors[0]):
        ndim -= 1

    for tensor in tensors:
        assert (
            tensor.ndim - (1 if torch.is_floating_point(tensor) else 0) == ndim
        )

    # Find the maximum shape
    max_shape = [
        max([tensor.shape[i] for tensor in tensors]) for i in range(ndim)
    ]

    # Expand x and y to the maximum shape
    expanded_tensors = []
    for tensor in tensors:
        expanded_tensors.append(
            tensor.expand(
                max_shape + ([-1] if torch.is_floating_point(tensor) else [])
            )
        )

    return tuple(expanded_tensors)



from typing import Optional

import torch
import torch.nn.functional as F

from torch_mist.baseline import ConstantBaseline
from torch_mist.critic import Critic
from torch_mist.estimators.discriminative.base import (
    BaselineDiscriminativeMIEstimator,
)


class JS(BaselineDiscriminativeMIEstimator):
    def __init__(
        self,
        critic: Critic,
        neg_samples: int = 1,
    ):
        super().__init__(
            critic=critic,
            neg_samples=neg_samples,
            baseline=ConstantBaseline(value=0.0),
        )

    def batch_loss(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
    ) -> Optional[torch.Tensor]:
        # Compute the critic on the positives. It has shape [...]
        f = self.unnormalized_log_ratio(x=x, y=y)

        x_, y_, log_w = self.sample_negatives(x, y)
        f_ = self.critic(x_, y_)

        pos = F.softplus(-f)
        neg = F.softplus(f_)

        # Compute the expectation w.r.t the M negatives (re-weighting if necessary)
        if not (log_w is None):
            neg = neg * log_w.exp()

        neg = neg.mean(0)

        assert pos.shape == neg.shape

        loss = pos + neg
        return loss



import os.path
from typing import Dict, Any, Optional

import pandas as pd

from torch_mist.utils.logging.logger.base import Logger


class PandasLogger(Logger):
    def __init__(
        self,
        log_dir: str = ".",
        log_name: Optional[str] = None,
        log_every: int = 1,
    ):
        super().__init__(log_dir=log_dir, log_every=log_every)
        self.__log = []
        if log_name is None:
            log_name = "log.csv"
        if not log_name.endswith(".csv"):
            log_name = log_name + ".csv"
        self.log_name = log_name

    def _log(
        self, data: Any, name: str, iteration: int, epoch: int, split: str
    ):
        if isinstance(data, Dict):
            for sub_name, value in data.items():
                self._log(
                    data=value,
                    name=f"{name}/{sub_name}",
                    iteration=iteration,
                    epoch=epoch,
                    split=split,
                )
        else:
            data = {
                "value": data,
                "name": name,
                "split": split,
                "iteration": iteration,
                "epoch": epoch,
            }
            self.__log.append(data)

    def get_log(self) -> pd.DataFrame:
        return pd.DataFrame(self.__log)

    def _reset_log(self):
        self.__log = []

    def _save_log(self, log: pd.DataFrame):
        log_path = os.path.join(self.log_dir, self.log_name)
        log.to_csv(log_path, index=False)



from typing import Dict, Union

import torch


def compute_mean_std(input, output) -> Dict[str, float]:
    assert isinstance(output, torch.Tensor) or isinstance(output, dict)
    if isinstance(output, dict):
        summary = {}
        for name, value in output.items():
            name = str(name)
            assert isinstance(value, torch.Tensor)
            summary[f"{name}/mean"] = torch.mean(value).item()
            summary[f"{name}/std"] = torch.std(value).item()
        return summary
    else:
        return {
            "mean": torch.mean(output).item(),
            "std": torch.std(output).item(),
        }


def compute_mean(input, output) -> Union[float, Dict[str, float]]:
    assert isinstance(output, torch.Tensor) or isinstance(output, dict)
    if isinstance(output, dict):
        summary = {}
        for name, value in output.items():
            name = str(name)
            assert isinstance(value, torch.Tensor)
            summary[name] = value.mean().item()
        return summary
    else:
        return output.mean().item()



from typing import Optional, Union, List

import torch
from pyro.distributions import ConditionalDistribution
from torch.distributions import Distribution

from torch_mist.distributions.factories import joint_transformed_normal
from torch_mist.distributions.joint.base import JointDistribution
from torch_mist.estimators.generative.implementations import (
    BA,
    CLUB,
    DoE,
    GM,
    L1Out,
    DummyGenerativeMIEstimator,
)


def ba(
    entropy_y: Union[float, torch.Tensor],
    x_dim: Optional[int] = None,
    y_dim: Optional[int] = None,
    hidden_dims: Optional[List[int]] = None,
    q_Y_given_X: Optional[ConditionalDistribution] = None,
    transform_name: str = "conditional_linear",
    n_transforms: int = 1,
) -> BA:
    from torch_mist.distributions.factories import (
        conditional_transformed_normal,
    )

    if q_Y_given_X is None:
        if x_dim is None or y_dim is None or hidden_dims is None:
            raise ValueError(
                "Either q_Y_given_X or x_dim, y_dim and hidden_dims must be provided"
            )

        q_Y_given_X = conditional_transformed_normal(
            input_dim=y_dim,
            context_dim=x_dim,
            hidden_dims=hidden_dims,
            transform_name=transform_name,
            n_transforms=n_transforms,
        )

    if isinstance(entropy_y, float):
        entropy_y = torch.FloatTensor([entropy_y])

    entropy_y = entropy_y.squeeze()

    return BA(
        q_Y_given_X=q_Y_given_X,
        entropy_y=entropy_y,
    )


def club(
    x_dim: Optional[int] = None,
    y_dim: Optional[int] = None,
    hidden_dims: Optional[List[int]] = None,
    q_Y_given_X: Optional[ConditionalDistribution] = None,
    transform_name: str = "conditional_linear",
    n_transforms: int = 1,
) -> CLUB:
    from torch_mist.distributions.factories import (
        conditional_transformed_normal,
    )

    if q_Y_given_X is None:
        if x_dim is None or y_dim is None or hidden_dims is None:
            raise ValueError(
                "Either q_Y_given_X or x_dim, y_dim and hidden_dims must be provided"
            )

        q_Y_given_X = conditional_transformed_normal(
            input_dim=y_dim,
            context_dim=x_dim,
            hidden_dims=hidden_dims,
            transform_name=transform_name,
            n_transforms=n_transforms,
        )

    return CLUB(
        q_Y_given_X=q_Y_given_X,
    )


def doe(
    x_dim: Optional[int] = None,
    y_dim: Optional[int] = None,
    hidden_dims: Optional[List[int]] = None,
    q_Y_given_X: Optional[ConditionalDistribution] = None,
    q_Y: Optional[Distribution] = None,
    conditional_transform_name: str = "conditional_linear",
    n_conditional_transforms: int = 1,
    marginal_transform_name: str = "linear",
    n_marginal_transforms: int = 1,
) -> DoE:
    from torch_mist.distributions.factories import (
        conditional_transformed_normal,
        transformed_normal,
    )

    if q_Y_given_X is None:
        if x_dim is None or y_dim is None or hidden_dims is None:
            raise ValueError(
                "Either q_Y_given_X or x_dim, y_dim and hidden_dims must be specified."
            )
        q_Y_given_X = conditional_transformed_normal(
            input_dim=y_dim,
            context_dim=x_dim,
            hidden_dims=hidden_dims,
            transform_name=conditional_transform_name,
            n_transforms=n_conditional_transforms,
        )

    if q_Y is None:
        if y_dim is None:
            raise ValueError(
                "Either q_Y or y_dim and hidden_dims must be specified."
            )
        q_Y = transformed_normal(
            input_dim=y_dim,
            hidden_dims=hidden_dims,
            transform_name=marginal_transform_name,
            n_transforms=n_marginal_transforms,
        )

    return DoE(
        q_Y_given_X=q_Y_given_X,
        q_Y=q_Y,
    )


def dummy_generative() -> DummyGenerativeMIEstimator:
    return DummyGenerativeMIEstimator()


def gm(
    x_dim: Optional[int] = None,
    y_dim: Optional[int] = None,
    hidden_dims: List[int] = None,
    q_XY: Optional[JointDistribution] = None,
    q_Y: Optional[Distribution] = None,
    q_X: Optional[Distribution] = None,
    joint_transform_name: str = "affine_autoregressive",
    n_joint_transforms: int = 1,
    marginal_transform_name: str = "linear",
    n_marginal_transforms: int = 1,
) -> GM:
    from torch_mist.distributions.factories import transformed_normal

    if q_XY is None:
        if x_dim is None or y_dim is None or hidden_dims is None:
            raise ValueError(
                "x_dim, y_dim, and hidden_dims must be provided if q_XY is not provided"
            )
        q_XY = joint_transformed_normal(
            input_dims={"x": x_dim, "y": y_dim},
            transform_name=joint_transform_name,
            n_transforms=n_joint_transforms,
        )

    if q_X is None:
        if x_dim is None or hidden_dims is None:
            raise ValueError(
                "x_dim and hidden_dims must be provided if q_X is not provided"
            )
        q_X = transformed_normal(
            input_dim=x_dim,
            hidden_dims=hidden_dims,
            transform_name=marginal_transform_name,
            n_transforms=n_marginal_transforms,
        )

    if q_Y is None:
        if y_dim is None or hidden_dims is None:
            raise ValueError(
                "y_dim and hidden_dims must be provided if q_Y is not provided"
            )
        q_Y = transformed_normal(
            input_dim=y_dim,
            hidden_dims=hidden_dims,
            transform_name=marginal_transform_name,
            n_transforms=n_marginal_transforms,
        )

    return GM(
        q_XY=q_XY,
        q_X=q_X,
        q_Y=q_Y,
    )


def l1out(
    x_dim: Optional[int] = None,
    y_dim: Optional[int] = None,
    hidden_dims: Optional[List[int]] = None,
    q_Y_given_X: Optional[ConditionalDistribution] = None,
    transform_name: str = "conditional_linear",
    n_transforms: int = 1,
) -> L1Out:
    from torch_mist.distributions.factories import (
        conditional_transformed_normal,
    )

    if q_Y_given_X is None:
        if x_dim is None or y_dim is None or hidden_dims is None:
            raise ValueError(
                "Either q_Y_given_X or x_dim, y_dim and hidden_dims must be provided"
            )

        q_Y_given_X = conditional_transformed_normal(
            input_dim=y_dim,
            context_dim=x_dim,
            hidden_dims=hidden_dims,
            transform_name=transform_name,
            n_transforms=n_transforms,
        )

    return L1Out(
        q_Y_given_X=q_Y_given_X,
    )



from typing import Dict, Any, Union, Optional

import numpy as np
import torch
from torch import nn

from torch_mist.decomposition.base import (
    DimensionalityReduction,
    StochasticDimensionalityReduction,
)
from torch_mist.distributions import (
    NormalModule,
    conditional_transformed_normal,
)
from torch_mist.estimators import instantiate_estimator
import torch_mist.models as tmm
from torch_mist.nn import dense_nn
from torch_mist.utils.data.utils import TensorDictLike
from torch_mist.utils.train import train_model

DEFAULT_MAX_ITERATIONS = 5000
DEFAULT_BATCH_SIZE = 64


class MIB(StochasticDimensionalityReduction):
    def __init__(self, *args, beta: float = 0.01, **kwargs):
        super().__init__(*args, **kwargs)
        self.beta = beta

    def _add_default_mi_estimator_params(
        self, model_params: Optional[Dict[str, Any]]
    ) -> Dict[str, Any]:
        if not ("hidden_dims" in model_params):
            model_params["hidden_dims"] = [128, 64]

        if not ("estimator_name" in model_params):
            model_params["estimator_name"] = "infonce"
            if not ("k_dim" in model_params):
                model_params["k_dim"] = 64
        return model_params

    def _add_default_model_params(
        self, model_params: Optional[Dict[str, Any]]
    ) -> Dict[str, Any]:
        if model_params is None:
            model_params = {}

        if not ("mi_estimator" in model_params):
            model_params["mi_estimator"] = {}

        model_params["mi_estimator"] = self._add_default_mi_estimator_params(
            model_params["mi_estimator"]
        )

        return model_params

    def _add_default_proj_params(
        self, proj_params: Optional[Dict[str, Any]]
    ) -> Dict[str, Any]:
        if proj_params is None:
            proj_params = {}
        if not ("hidden_dims" in proj_params):
            proj_params["hidden_dims"] = [128]
        if not ("nonlinearity" in proj_params):
            proj_params["nonlinearity"] = nn.ReLU(True)

        return proj_params

    def _add_default_y_proj_params(
        self, proj_params: Optional[Dict[str, Any]]
    ) -> Dict[str, Any]:
        if proj_params is None:
            proj_params = {}
        if not ("hidden_dims" in proj_params):
            proj_params["hidden_dims"] = [128]
        if not ("nonlinearity" in proj_params):
            proj_params["nonlinearity"] = nn.ReLU(True)

        return proj_params

    def _instantiate_y_proj(self, y_dim: int) -> nn.Module:
        self.proj_params["input_dim"] = self.n_dim
        self.proj_params["context_dim"] = y_dim
        return conditional_transformed_normal(**self.proj_params)

    def _train_model(self, data: TensorDictLike, **train_params):
        return train_model(
            self.model,
            data,
            eval_method="loss",
            train_logged_methods=["regularization", "loss"],
            eval_logged_methods=["regularization", "mutual_information"],
            **train_params
        )

    def _instantiate_model(self, x_dim: int, y_dim: int):
        self.model_params["mi_estimator"]["x_dim"] = self.n_dim
        self.model_params["mi_estimator"]["y_dim"] = self.n_dim
        self.model_params["mi_estimator"] = instantiate_estimator(
            **self.model_params["mi_estimator"]
        )

        return tmm.bottleneck.MIB(
            p_ZX_given_X=self.proj,
            p_ZY_given_Y=self.y_proj,
            beta=self.beta,
            **self.model_params
        )



import os
import tempfile
from typing import Dict, Any, Optional
import torch
from torch import nn

import wandb
from torch_mist.utils.logging.logger.base import Logger


class WandbLogger(Logger):
    def __init__(
        self, project: str, log_dir: Optional[str] = None, log_every: int = 10
    ):
        if log_dir is None:
            log_dir = tempfile.gettempdir()
        super().__init__(log_dir=log_dir, log_every=log_every)
        self.run = wandb.init(project=project)
        self.run.define_metric("epoch", hidden=True)
        self.run.define_metric("iteration", hidden=True)
        self.run.define_metric("train/*", step_metric="iteration")
        self.run.define_metric("valid/*", step_metric="epoch")
        self.project = project

    def add_config(self, config: Dict[str, Any]):
        self.run.config.update(config)

    def _log(
        self, data: Any, name: str, iteration: int, epoch: int, split: str
    ):
        name = name.replace(".", "/")
        if isinstance(data, dict):
            entry = {
                f"{name}/{key.replace('.','/')}": value
                for key, value in data.items()
            }
        else:
            entry = {name: data}

        entry = {f"{split}/{name}": value for name, value in entry.items()}

        entry["epoch"] = epoch
        entry["iteration"] = iteration

        wandb.log(entry, step=iteration)

    def get_log(self) -> wandb.wandb_sdk.wandb_run.Run:
        return self.run

    def _reset_log(self):
        self.run = wandb.init(project=self.project)

    def save_model(
        self, model: nn.Module, name: str, artifact_name: Optional[str] = None
    ):
        filepath = os.path.join(self.log_dir, name)
        torch.save(model, filepath)

        if artifact_name is None:
            artifact_name = name
        artifact = wandb.Artifact(artifact_name, type="model")
        artifact.add_file(filepath)
        self.run.log_artifact(artifact)



from typing import Tuple, Optional, Callable

import torch
from pyro.distributions import ConditionalDistribution

from torch_mist.estimators.discriminative.base import DiscriminativeMIEstimator
from torch_mist.estimators.hybrid.base import HybridMIEstimator
from torch_mist.estimators.transformed.implementations.pq import PQ
from torch_mist.quantization import QuantizationFunction, LearnableQuantization
from torch_mist.utils.caching import cached_method


class PQHybridMIEstimator(HybridMIEstimator):
    def __init__(
        self,
        q_QY_given_X: ConditionalDistribution,
        quantize_y: QuantizationFunction,
        discriminative_estimator: DiscriminativeMIEstimator,
        temperature: float = 1.0,
    ):
        pq_estimator = PQ(
            q_QY_given_X=q_QY_given_X,
            quantize_y=quantize_y,
            temperature=temperature,
        )

        super().__init__(
            generative_estimator=pq_estimator,
            discriminative_estimator=discriminative_estimator,
        )
        self._validate_batches = True

    def disable_batch_validation(self):
        self._validate_batches = False

    def enable_batch_validation(self):
        self._validate_batches = True

    @property
    def quantize_y(self) -> Callable[[torch.Tensor], torch.LongTensor]:
        return self.generative_estimator.transforms["y->y"]

    @cached_method
    def sample_negatives(
        self, x: torch.Tensor, y: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, Optional[torch.Tensor]]:
        # We reshape the batches from [batch_size, ...] to [neg_samples+1, batch_size//neg_samples, ...]
        # This is because the batches are created so that the negatives can be sampled by shuffling the new first dim
        if self.neg_samples <= 0:
            neg_shape = x.shape[0]
        else:
            neg_shape = self.neg_samples + 1

        batch_shape = x.shape[0] // (neg_shape)

        # If the batch validation is disabled, we skip checking the batches to speed up computation
        if self._validate_batches:
            Q_y = self.quantize_y(y)
            Q_y = Q_y.view(neg_shape, batch_shape, *Q_y.shape[1:])

            # Check the labels are the same for the corresponding ys, (this can be done using the SameAttributeSampler)
            Q_y0 = Q_y[0]
            assert (
                torch.sum(Q_y0.unsqueeze(0) == Q_y).item() == Q_y.numel()
            ), Q_y

        # Sample negatives as usual by shuffling the (new) first dimension
        # this way we shuffle only y with the same Q_y
        x_, y_, log_w = DiscriminativeMIEstimator.sample_negatives(
            self,
            x.view(neg_shape, batch_shape, *x.shape[1:]),
            y.view(neg_shape, batch_shape, *y.shape[1:]),
        )

        x_ = x.unsqueeze(0)
        # Collapse the intermediate dimensions to obtain [neg_samples, batch_size, ...]
        y_ = y_.reshape(y_.shape[0], -1, *y_.shape[3:])
        if log_w:
            log_w = log_w.reshape(log_w.shape[0], -1, *log_w.shape[3:])

        return x_, y_, log_w



from typing import List, Dict

import numpy as np
import torch
from pyro.distributions import Independent
from torch.distributions import (
    MultivariateNormal,
    MixtureSameFamily,
    Categorical,
    Normal,
)

from torch_mist.distributions.joint.base import JointDistribution
from torch_mist.distributions.joint.wrapper import TorchJointDistribution


class MultivariateCorrelatedNormalMixture(TorchJointDistribution):
    MC_SAMPLES = 100000
    MC_ITERATIONS = 100

    def __init__(
        self,
        rho: float = 0.95,
        sigma: float = 0.1,
        epsilon: float = 0.15,
        delta: float = 1.5,
        n_dim: int = 5,
    ):
        self.n_dim = n_dim
        self.rho = rho
        self.sigma = sigma
        self.epsilon = epsilon
        self.delta = delta

        covariance = torch.eye(2)
        covariance[0, 1] = covariance[1, 0] = rho
        covariance = covariance * sigma

        mu = torch.zeros(n_dim, 4, 2)

        mu[:, 0, 0] = epsilon + delta
        mu[:, 0, 1] = -epsilon + delta
        mu[:, 1, 0] = -epsilon - delta
        mu[:, 1, 1] = epsilon - delta
        mu[:, 2, 0] = epsilon - delta
        mu[:, 2, 1] = -epsilon - delta
        mu[:, 3, 0] = -epsilon + delta
        mu[:, 3, 1] = epsilon + delta

        self.x_component_dist = Normal(mu[:, :, 0], sigma**0.5)

        # Store the marginal distribution for one dimension
        self.p_X = Independent(
            MixtureSameFamily(
                Categorical(logits=torch.zeros(n_dim, 4)),
                self.x_component_dist,
            ),
            1,
        )

        self.component_dist = MultivariateNormal(
            mu, covariance.unsqueeze(0).unsqueeze(1)
        )

        p_XY = Independent(
            MixtureSameFamily(
                Categorical(logits=torch.zeros(n_dim, 4)),
                self.component_dist,
            ),
            1,
        )

        self._cached_estimates = {}

        super().__init__(torch_dist=p_XY, variables=["x", "y"])

    def _marginal(self, variables: List[str]) -> JointDistribution:
        assert len(variables) == 1
        variable = variables[0]

        return TorchJointDistribution(
            torch_dist=self.p_X,
            variables=[variable],
        )

    def _approximate_estimates(self):
        if self.n_dim > 1:
            p_XY = MultivariateCorrelatedNormalMixture(
                n_dim=1,
                rho=self.rho,
                delta=self.delta,
                sigma=self.sigma,
                epsilon=self.epsilon,
            )
        else:
            p_XY = self

        estimates = {
            "H(x)": [],
            "H(y)": [],
            "H(xy)": [],
            "I(x;y)": [],
        }

        p_X = p_XY.marginal("x")
        p_Y = p_XY.marginal("y")

        for i in range(self.MC_ITERATIONS):
            # Sample from the joint
            samples = p_XY.sample([self.MC_SAMPLES])

            # Compute the log-probability of each sample pair
            log_p_xy = p_XY.log_prob(**samples)

            # And the marginal log-probability
            log_p_x = p_X.log_prob(samples["x"])
            log_p_y = p_Y.log_prob(samples["y"])

            # MC estimation of the joint and marginal entropies
            entropy_xy = -torch.mean(log_p_xy)
            entropy_x = -torch.mean(log_p_x)
            entropy_y = -torch.mean(log_p_y)

            estimates["H(x)"].append(entropy_x)
            estimates["H(x)"].append(entropy_y)
            estimates["H(xy)"].append(entropy_xy)
            estimates["I(x;y)"].append(entropy_x + entropy_y - entropy_xy)

        for name, values in estimates.items():
            self._cached_estimates[name] = np.mean(values) * self.n_dim

    def _entropy(self, variables: List[str]) -> torch.Tensor:
        if len(self._cached_estimates) == 0:
            self._approximate_estimates()
        if len(variables) == 2:
            return self._cached_estimates["H(xy)"]

        return self._cached_estimates["H(x)"]

    def _mutual_information(
        self, variable_1: str, variable_2: str
    ) -> torch.Tensor:
        if len(self._cached_estimates) == 0:
            self._approximate_estimates()
        return self._cached_estimates["I(x;y)"]



from typing import List

import torch
from pyro.nn import DenseNN
from torch import nn

from .base import Critic
from torch_mist.utils.shape import expand_to_same_shape


class JointCritic(Critic):
    def __init__(
        self,
        joint_net: nn.Module,
    ):
        """
        Model the critic as a function of (x,y).
        :param joint_net: a (learnable) model returning a real number when given pairs (x,y)
        """

        super(JointCritic, self).__init__()
        self.joint_net = joint_net

    def forward(self, x, y) -> torch.Tensor:
        x, y = expand_to_same_shape(x, y)

        # Concatenate x and y in order
        xy = torch.cat([x, y], -1)

        return self.joint_net(xy).squeeze(-1)



import torch
from pyro.distributions import ConditionalDistribution
from torch.distributions import Distribution


class UnconditionalDistribution(ConditionalDistribution):
    def __init__(self, marginal_distribution: Distribution):
        super().__init__()
        self.marginal_distribution = marginal_distribution

    def condition(self, context: torch.Tensor) -> Distribution:
        return self.marginal_distribution

    def __repr__(self):
        s = f"{self.__class__.__name__}(\n"
        s += (
            "    (marginal_distribution): "
            + self.marginal_distribution.__repr__().replace("\n", "\n  ")
        )
        s += "  \n)"
        return s



from typing import Union, Dict

import torch

from torch_mist.distributions import JointDistribution


def prepare_samples(
    distribution: JointDistribution, n_samples: int
) -> Union[torch.Tensor, Dict[str, torch.Tensor]]:
    return distribution.sample(torch.Size([n_samples]))


def compute_mi(distribution: JointDistribution, **kwargs) -> float:
    return distribution.mutual_information(**kwargs).item()



import torch
from torch_mist.utils.caching import cached_function
from typing import Tuple


@cached_function
def _make_indexing_mask(n: int, k: int, device: torch.Tensor) -> torch.Tensor:
    # Consider the first K off-diagonal elements
    idx = torch.arange(n * k).to(device).view(n, k).long()
    idx = (idx % k + torch.div(idx, k, rounding_mode="floor") + 1) % n
    return idx.T


def select_k_others(x: torch.Tensor, k: int) -> torch.Tensor:
    N = x.shape[0]
    idx = _make_indexing_mask(N, k, x.device)
    return x[idx]


@cached_function
def _make_off_selection_indexing_mask(
    n: int, k: int, device: torch.device
) -> Tuple[torch.Tensor, torch.Tensor]:
    cols = torch.arange(n).unsqueeze(0).repeat(k, 1)
    rows = (cols + 1 + torch.arange(k).unsqueeze(1)) % n
    return rows.to(device), cols.to(device)


def matrix_off_diagonal(matrix: torch.Tensor, k: int):
    N = matrix.shape[0]
    rows, cols = _make_off_selection_indexing_mask(N, k, matrix.device)
    res = matrix[rows, cols, ...]
    return res



from typing import List, Optional, Callable

from sklearn.cluster import KMeans

from torch_mist.distributions.factories import conditional_transformed_normal
from torch_mist.nn import dense_nn
from torch_mist.quantization.functions import (
    ClusterQuantization,
    QuantizationFunction,
)
from torch_mist.models.vqvae import VQVAE


def kmeans_quantization(
    n_bins: int, n_init="auto", **kwargs
) -> ClusterQuantization:
    clustering = KMeans(n_bins, n_init=n_init, **kwargs)
    return ClusterQuantization(clustering)


def vqvae(
    input_dim: int,
    quantization_dim: int,
    n_bins: int,
    hidden_dims: List[int],
    beta: float = 0.2,
    nonlinearity: Optional[Callable] = None,
    version: str = "v2",
    **train_params,
) -> VQVAE:
    assert len(hidden_dims) > 0, "hidden_dims must be a non-empty list"

    if quantization_dim is None:
        quantization_dim = 16

    encoder = dense_nn(
        input_dim,
        output_dim=quantization_dim,
        hidden_dims=hidden_dims,
        nonlinearity=nonlinearity,
    )
    decoder = conditional_transformed_normal(
        input_dim=input_dim,
        context_dim=quantization_dim,
        transform_name="conditional_linear",
        hidden_dims=hidden_dims,
    )

    return VQVAE(
        encoder=encoder,
        decoder=decoder,
        n_bins=n_bins,
        quantization_dim=quantization_dim,
        beta=beta,
        version=version,
        **train_params,
    )


def instantiate_quantization(
    name: str, n_bins: int, **kwargs
) -> QuantizationFunction:
    if name == "kmeans":
        quantization = kmeans_quantization(n_bins=n_bins, **kwargs)
    elif name == "vqvae":
        quantization = vqvae(n_bins=n_bins, **kwargs)
    else:
        raise ValueError(f"The quantization scheme {name} is not supported.")

    return quantization



from .base import (
    DistributionModule,
    ConditionalDistributionModule,
    TransformedDistributionModule,
    ConditionalTransformedDistributionModule,
)
from .implementations import *
from .factories import *
from .utils import fetch_transform



from typing import Optional, List

import torch
from torch import nn
from pyro.nn import DenseNN

from torch_mist.distributions.transforms.implementations.permute import Permute
from torch_mist.distributions.transforms.implementations.linear import (
    Linear,
    ConditionalLinear,
)
from torch_mist.distributions.transforms.implementations.normalize import (
    EMANormalize,
)


def linear(
    input_dim: int,
    loc: Optional[float] = None,
    scale: Optional[float] = None,
    initial_scale: Optional[float] = None,
) -> Linear:
    return Linear(input_dim, scale=scale, initial_scale=initial_scale, loc=loc)


def conditional_linear(
    input_dim: int,
    context_dim: int,
    hidden_dims: Optional[List[int]] = None,
    scale: Optional[float] = None,
    initial_scale: Optional[float] = None,
    nonlinearity: nn.Module = nn.ReLU(True),
) -> ConditionalLinear:
    if hidden_dims is None:
        hidden_dims = [input_dim * 10, input_dim * 10]
    if scale is None:
        nn = DenseNN(
            input_dim=context_dim,
            hidden_dims=hidden_dims,
            param_dims=[input_dim, input_dim],
            nonlinearity=nonlinearity,
        )
    else:
        nn = DenseNN(
            context_dim,
            hidden_dims,
            param_dims=[input_dim],
            nonlinearity=nonlinearity,
        )
    return ConditionalLinear(nn, scale=scale, initial_scale=initial_scale)


def permute(
    input_dim: int, permutation: Optional[List[int]] = None, dim: int = -1
) -> Permute:
    """
    A helper function to create a :class:`~pyro.distributions.transforms.Permute`
    object for consistency with other helpers.

    :param input_dim: Dimension(s) of input variable to permute. Note that when
        `dim < -1` this must be a tuple corresponding to the event shape.
    :type input_dim: int
    :param permutation: Torch tensor of integer indices representing permutation.
        Defaults to a random permutation.
    :type permutation: torch.LongTensor
    :param dim: the tensor dimension to permute. This value must be negative and
        defines the event dim as `abs(dim)`.
    :type dim: int

    """
    if dim < -1 or not isinstance(input_dim, int):
        if len(input_dim) != -dim:
            raise ValueError(
                "event shape {} must have same length as event_dim {}".format(
                    input_dim, -dim
                )
            )
        input_dim = input_dim[dim]

    if permutation is None:
        permutation = torch.randperm(input_dim)
    return Permute(permutation, dim=dim)


def emanorm(
    input_dim: int, gamma: float = 0.99, normalize_inverse: bool = True
) -> EMANormalize:
    return EMANormalize(
        input_dim=input_dim, gamma=gamma, normalize_inverse=normalize_inverse
    )



from abc import abstractmethod
from typing import Optional, Union

import torch
from pyro.distributions import ConditionalDistribution
from torch import nn
from torch.distributions import Distribution

from torch_mist.distributions.conditional.delta import ConditionalDelta
from torch_mist.estimators import MIEstimator
from torch_mist.nn import Identity, Model


class InformationBottleneck(Model):
    def __init__(
        self,
        mi_estimator: MIEstimator,
        beta: float,
        p_ZX_given_X: Optional[
            Union[nn.Module, ConditionalDistribution]
        ] = None,
        p_ZY_given_Y: Optional[
            Union[nn.Module, ConditionalDistribution]
        ] = None,
    ):
        super().__init__()
        self.upper_bound = mi_estimator.lower_bound

        self.beta = beta
        self.mi_estimator = mi_estimator

        if p_ZX_given_X is None:
            p_ZX_given_X = Identity()
        if p_ZY_given_Y is None:
            p_ZY_given_Y = Identity()

        if not isinstance(p_ZX_given_X, ConditionalDistribution):
            p_ZX_given_X = ConditionalDelta(p_ZX_given_X)
        if not isinstance(p_ZY_given_Y, ConditionalDistribution):
            p_ZY_given_Y = ConditionalDelta(p_ZY_given_Y)

        self.encoder_x = p_ZX_given_X
        self.encoder_y = p_ZY_given_Y

    @abstractmethod
    def regularization(
        self,
        zx: torch.Tensor,
        zy: torch.Tensor,
        p_ZX_given_x: Distribution,
        p_ZY_given_y: Distribution,
    ) -> torch.Tensor():
        raise NotImplementedError()

    def mutual_information(self, x: torch.Tensor, y: torch.Tensor):
        return self.mi_estimator.mutual_information(x, y)

    def loss(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        p_zx_given_x = self.encoder_x.condition(x)
        p_zy_given_y = self.encoder_y.condition(y)
        zx = p_zx_given_x.rsample()
        zy = p_zy_given_y.rsample()

        mi_loss = self.mi_estimator.loss(zx, zy)
        reg_loss = self.regularization(zx, zy, p_zx_given_x, p_zy_given_y)
        assert reg_loss.numel() == 1

        return mi_loss + self.beta * reg_loss


class VIB(InformationBottleneck):
    def __init__(
        self,
        mi_estimator: MIEstimator,
        q_ZX: Distribution,
        beta: float,
        p_ZX_given_X: Optional[
            Union[nn.Module, ConditionalDistribution]
        ] = None,
        p_ZY_given_Y: Optional[
            Union[nn.Module, ConditionalDistribution]
        ] = None,
    ):
        super().__init__(
            mi_estimator=mi_estimator,
            beta=beta,
            p_ZX_given_X=p_ZX_given_X,
            p_ZY_given_Y=p_ZY_given_Y,
        )
        self.q_ZX = q_ZX

    def regularization(
        self,
        zx: torch.Tensor,
        zy: torch.Tensor,
        p_ZX_given_x: Distribution,
        p_ZY_given_y: Distribution,
    ) -> torch.Tensor():
        KL = (p_ZX_given_x.log_prob(zx) - self.q_ZX.log_prob(zx)).mean()

        return KL


class MIB(InformationBottleneck):
    def regularization(
        self,
        zx: torch.Tensor,
        zy: torch.Tensor,
        p_ZX_given_x: Distribution,
        p_ZY_given_y: Distribution,
    ) -> torch.Tensor():
        assert (
            zx.shape == zy.shape
        ), f"MIB requires representations of the same shape, {zx.shape}!={zy.shape}"

        KL_x = p_ZX_given_x.log_prob(zx) - p_ZY_given_y.log_prob(zx)
        KL_y = p_ZY_given_y.log_prob(zy) - p_ZX_given_x.log_prob(zy)

        return (KL_x + KL_y).mean() / 2.0


class CEB(InformationBottleneck):
    def __init__(
        self,
        mi_estimator: MIEstimator,
        q_ZX_given_ZY: ConditionalDistribution,
        beta: float,
        p_ZX_given_X: Optional[
            Union[nn.Module, ConditionalDistribution]
        ] = None,
        p_ZY_given_Y: Optional[
            Union[nn.Module, ConditionalDistribution]
        ] = None,
    ):
        super().__init__(
            mi_estimator=mi_estimator,
            beta=beta,
            p_ZX_given_X=p_ZX_given_X,
            p_ZY_given_Y=p_ZY_given_Y,
        )
        self.q_ZX_given_ZY = q_ZX_given_ZY

    def regularization(
        self,
        zx: torch.Tensor,
        zy: torch.Tensor,
        p_ZX_given_x: Distribution,
        p_ZY_given_y: Distribution,
    ) -> torch.Tensor():
        q_ZX_given_zy = self.q_ZX_given_ZY.condition(zy)
        KL = (p_ZX_given_x.log_prob(zx) - q_ZX_given_zy.log_prob(zx)).mean()

        return KL



from typing import Dict, Union, Callable, Any

import numpy as np
import torch
from torch.utils.data import (
    DataLoader,
    Dataset,
)

from torch_mist.utils.data.sampler import SameAttributeSampler


class SameAttributeDataLoader(DataLoader):
    def __init__(
        self,
        dataset: Dataset,
        attributes: Union[torch.Tensor, np.ndarray],
        batch_size: int,
        neg_samples: int,
        **kwargs
    ):
        if len(dataset) != len(attributes):
            raise Exception(
                "The dataset and the attributes must have the same length."
            )

        super().__init__(
            dataset,
            batch_sampler=SameAttributeSampler(
                attributes=attributes,
                batch_size=batch_size,
                neg_samples=neg_samples,
            ),
            **kwargs
        )


def sample_same_attributes(
    dataloader: DataLoader,
    attributes: Union[torch.Tensor, np.ndarray],
    neg_samples: int,
) -> DataLoader:
    return SameAttributeDataLoader(
        neg_samples=neg_samples,
        dataset=dataloader.dataset,
        attributes=attributes,
        batch_size=dataloader.batch_size,
        num_workers=dataloader.num_workers,
        prefetch_factor=dataloader.prefetch_factor,
        pin_memory=dataloader.pin_memory,
        pin_memory_device=dataloader.pin_memory_device,
        timeout=dataloader.timeout,
        worker_init_fn=dataloader.worker_init_fn,
        multiprocessing_context=dataloader.multiprocessing_context,
    )


def sample_same_value(
    dataloader: DataLoader,
    func: Callable[[Any], Any],
    neg_samples: int,
) -> DataLoader:
    ordered_dataloader = DataLoader(
        dataset=dataloader.dataset,
        batch_size=dataloader.batch_size,
        num_workers=dataloader.num_workers,
        prefetch_factor=dataloader.prefetch_factor,
        pin_memory=dataloader.pin_memory,
        pin_memory_device=dataloader.pin_memory_device,
        timeout=dataloader.timeout,
        worker_init_fn=dataloader.worker_init_fn,
        multiprocessing_context=dataloader.multiprocessing_context,
        shuffle=False,
    )

    attributes = []
    for batch in ordered_dataloader:
        attributes.append(func(batch))

    attributes = torch.cat(attributes, 0)

    return sample_same_attributes(
        dataloader, attributes, neg_samples=neg_samples
    )



from abc import abstractmethod
from typing import Optional, Dict, Any, Union

import numpy as np
from pyro.distributions import ConditionalDistribution
from sklearn.base import TransformerMixin
import torch
from torch import nn

from torch_mist.distributions import conditional_transformed_normal
from torch_mist.estimators import (
    MIEstimator,
    instantiate_estimator,
    TransformedMIEstimator,
)
from torch_mist.nn import dense_nn, Model
from torch_mist.utils import train_mi_estimator
from torch_mist.utils.data.utils import TensorDictLike
from torch_mist.utils.train import train_model

DEFAULT_MAX_ITERATIONS = 5000
DEFAULT_BATCH_SIZE = 64


class CenterAndScale:
    def __init__(
        self,
        loc: Union[torch.Tensor, np.ndarray],
        scale: Union[torch.Tensor, np.ndarray],
        min_scale=1e-6,
    ):
        super().__init__()
        self.loc = loc
        scale[np.abs(scale) < min_scale] = 1.0
        self.scale = scale

    def __call__(self, data):
        return (data - self.loc) / self.scale


class DimensionalityReduction(TransformerMixin):
    def __init__(
        self,
        n_dim: int,
        normalize_inputs: bool = True,
        whiten: bool = False,
        proj: Optional[nn.Module] = None,
        y_proj: Optional[nn.Module] = None,
        model: Optional[Model] = None,
        proj_params: Optional[Dict[str, Any]] = None,
        y_proj_params: Optional[Dict[str, Any]] = None,
        model_params: Optional[Dict[str, Any]] = None,
    ):
        super().__init__()
        self.n_dim = n_dim
        self.proj = proj
        self.y_proj = y_proj
        self.model = model

        self.proj_params = self._add_default_proj_params(proj_params)
        self.y_proj_params = self._add_default_y_proj_params(y_proj_params)
        self.model_params = self._add_default_model_params(model_params)

        self.normalize_inputs = normalize_inputs
        self.normalize_projection = whiten
        self.normalize_X = None
        self.normalize_Y = None
        self.normalize_Z = None
        self.train_log = None

    def _add_default_model_params(
        self, model_params: Optional[Dict[str, Any]]
    ) -> Optional[Dict[str, Any]]:
        return model_params

    def _add_default_proj_params(
        self, proj_params: Optional[Dict[str, Any]]
    ) -> Optional[Dict[str, Any]]:
        return proj_params

    def _add_default_y_proj_params(
        self, y_proj_params: Optional[Dict[str, Any]]
    ) -> Optional[Dict[str, Any]]:
        return y_proj_params

    def _add_default_train_params(
        self, train_params: Optional[Dict[str, Any]]
    ) -> Dict[str, Any]:
        if train_params is None:
            train_params = {}

        if not ("batch_size" in train_params):
            print(
                "[Info]: batch_size is not specified,"
                + f" using batch_size={DEFAULT_BATCH_SIZE} by default."
            )
            train_params["batch_size"] = DEFAULT_BATCH_SIZE

        if not ("max_epochs" in train_params) and not (
            "max_iterations" in train_params
        ):
            print(
                "[Info]: max_epoch and max_iterations are not specified,"
                + f" using max_iterations={DEFAULT_MAX_ITERATIONS} by default."
            )
            train_params["max_iterations"] = DEFAULT_MAX_ITERATIONS

        if not ("early_stopping" in train_params):
            train_params["early_stopping"] = True

        if not ("verbose" in train_params):
            train_params["verbose"] = False

        return train_params

    @abstractmethod
    def _instantiate_proj(self, x_dim: int) -> nn.Module:
        raise NotImplementedError()

    @abstractmethod
    def _instantiate_y_proj(self, y_dim: int) -> Optional[nn.Module]:
        raise NotImplementedError()

    @abstractmethod
    def _instantiate_model(self, x_dim: int, y_dim: int) -> Model:
        raise NotImplementedError()

    def _train_model(self, data: TensorDictLike, **train_params):
        return train_model(
            self.model, data, eval_method="loss", **train_params
        )

    def _get_transformed_y_dim(self, Y: np.ndarray) -> int:
        if self.y_proj is None:
            y_dim = Y.shape[-1]
        else:
            y_dim = self.y_proj(torch.FloatTensor(Y[:2])).shape[-1]
        return y_dim

    def fit(
        self,
        X: Union[np.ndarray, torch.Tensor],
        Y: Union[np.ndarray, torch.Tensor],
        **train_params,
    ):
        if X.ndim == 1:
            X = X.reshape(-1, 1)
        if Y.ndim == 1:
            Y = Y.reshape(-1, 1)

        x_dim = X.shape[-1]
        y_dim = self._get_transformed_y_dim(Y)

        if self.normalize_inputs:
            self.normalize_X = CenterAndScale(
                X.mean(0, keepdims=True), X.std(0, keepdims=True)
            )
            CX = self.normalize_X(X)
            self.normalize_Y = CenterAndScale(
                Y.mean(0, keepdims=True), Y.std(0, keepdims=True)
            )
            CY = self.normalize_Y(Y)
        else:
            CX = X
            CY = Y

        if self.proj is None:
            self.proj = self._instantiate_proj(x_dim)
        if self.y_proj is None:
            self.y_proj = self._instantiate_y_proj(y_dim)
        if self.model is None:
            self.model = self._instantiate_model(x_dim=x_dim, y_dim=y_dim)

        train_params = self._add_default_train_params(train_params)

        self.train_log = self._train_model(data=(CX, CY), **train_params)

        if self.normalize_projection:
            self.normalize_Z = None
            Z = self.transform(X)
            self.normalize_Z = CenterAndScale(
                Z.mean(0, keepdims=True), Z.std(0, keepdims=True)
            )

        return self

    def _encode(self, X: torch.Tensor) -> torch.Tensor:
        return self.proj(X)

    def transform(
        self,
        X: np.ndarray,
    ) -> Union[np.ndarray, torch.Tensor]:
        if self.proj is None:
            raise ValueError(
                "Use the fit(X, Y) method before calling transform(X)."
            )
        self.proj.eval()

        if self.normalize_X:
            X = self.normalize_X(X)

        is_numpy = False
        device = next(iter(self.proj.parameters())).device

        if isinstance(X, np.ndarray):
            X = torch.FloatTensor(X)
            is_numpy = True

        original_device = X.device
        with torch.no_grad():
            z = self._encode(X.to(device)).to(original_device)
            if not (self.normalize_Z is None):
                z = self.normalize_Z(z)

        if is_numpy:
            z = z.data.numpy()

        return z


class StochasticDimensionalityReduction(DimensionalityReduction):
    def __init__(self, *args, stochastic_transform: bool = False, **kwargs):
        super().__init__(*args, **kwargs)
        self.stochastic_transform = stochastic_transform

    def _instantiate_proj(self, x_dim: int) -> nn.Module:
        self.proj_params["input_dim"] = self.n_dim
        self.proj_params["context_dim"] = x_dim
        return conditional_transformed_normal(**self.proj_params)

    def _get_transformed_y_dim(self, Y: np.ndarray) -> int:
        if self.y_proj is None:
            y_dim = Y.shape[-1]
        elif isinstance(self.y_proj, ConditionalDistribution):
            y_dim = (
                self.y_proj.condition(torch.FloatTensor(Y[:2]))
                .sample()
                .shape[-1]
            )
        else:
            y_dim = self.y_proj(torch.FloatTensor(Y[:2])).shape[-1]
        return y_dim

    def _encode(self, X: torch.Tensor) -> torch.Tensor:
        p_Z_given_x = self.proj.condition(X)
        if self.stochastic_transform:
            return p_Z_given_x.sample()
        else:
            z = torch.zeros(*X.shape[:-1], self.n_dim).to(X.device)
            # Transform the mean of the base distribution (which is also the mode for normals).
            z += p_Z_given_x.base_dist.mean
            for transform in p_Z_given_x.transforms:
                z = transform._call(z)
            return z



import inspect
from typing import List

from torch_mist.baseline.base import ConstantBaseline
from torch_mist.baseline.factories import baseline_nn
from torch_mist.critic.base import JOINT_CRITIC, SEPARABLE_CRITIC
from torch_mist.critic.factories import critic_nn, shared_critic_nns
from torch_mist.estimators.discriminative.implementations import (
    AlphaTUBA,
    FLO,
    InfoNCE,
    JS,
    MINE,
    NWJ,
    SMILE,
    TUBA,
    DummyDiscriminativeMIEstimator,
)


def alpha_tuba(
    x_dim: int,
    y_dim: int,
    hidden_dims: List[int],
    alpha: float = 0.01,
    learnable_baseline: bool = True,
    critic_type: str = SEPARABLE_CRITIC,
    neg_samples: int = 0,
    **kwargs,
) -> AlphaTUBA:
    baseline_params = {}
    critic_params = {}

    for param_name, param_value in kwargs.items():
        if param_name in inspect.signature(baseline_nn).parameters:
            baseline_params[param_name] = param_value
        else:
            critic_params[param_name] = param_value
        if param_name in inspect.signature(critic_nn).parameters:
            critic_params[param_name] = param_value

    if learnable_baseline:
        b_nn = baseline_nn(
            x_dim=x_dim, hidden_dims=hidden_dims, **baseline_params
        )
    else:
        b_nn = ConstantBaseline(value=1.0)

    return AlphaTUBA(
        critic=critic_nn(
            x_dim=x_dim,
            y_dim=y_dim,
            hidden_dims=hidden_dims,
            critic_type=critic_type,
            **critic_params,
        ),
        baseline=b_nn,
        alpha=alpha,
        neg_samples=neg_samples,
    )


def flo(
    x_dim: int,
    y_dim: int,
    hidden_dims: List[int],
    neg_samples: int = -1,
    n_shared_layers: int = -1,
    critic_type: str = SEPARABLE_CRITIC,
    **kwargs,
) -> FLO:
    # Make two critics with shared architectures
    critic, normalized_critic = shared_critic_nns(
        x_dim=x_dim,
        y_dim=y_dim,
        hidden_dims=hidden_dims,
        n_shared_layers=n_shared_layers,
        critic_type=critic_type,
        n_critics=2,
        **kwargs,
    )

    return FLO(
        critic=critic,
        normalized_critic=normalized_critic,
        neg_samples=neg_samples,
    )


def infonce(
    x_dim: int,
    y_dim: int,
    hidden_dims: List[int],
    critic_type: str = SEPARABLE_CRITIC,
    neg_samples: int = 0,
    **kwargs,
) -> InfoNCE:
    return InfoNCE(
        critic=critic_nn(
            x_dim=x_dim,
            y_dim=y_dim,
            critic_type=critic_type,
            hidden_dims=hidden_dims,
            **kwargs,
        ),
        neg_samples=neg_samples,
    )


def js(
    x_dim: int,
    y_dim: int,
    hidden_dims: List[int],
    neg_samples: int = 1,
    critic_type: str = JOINT_CRITIC,
    **kwargs,
) -> JS:
    return JS(
        critic=critic_nn(
            x_dim=x_dim,
            y_dim=y_dim,
            hidden_dims=hidden_dims,
            critic_type=critic_type,
            **kwargs,
        ),
        neg_samples=neg_samples,
    )


def dummy_discriminative(
    neg_samples: int = 1,
    **kwargs,
) -> DummyDiscriminativeMIEstimator:
    return DummyDiscriminativeMIEstimator(
        neg_samples=neg_samples,
    )


def mine(
    x_dim: int,
    y_dim: int,
    hidden_dims: List[int],
    critic_type: str = JOINT_CRITIC,
    neg_samples: int = 1,
    gamma: float = 0.9,
    **kwargs,
) -> MINE:
    return MINE(
        critic=critic_nn(
            x_dim=x_dim,
            y_dim=y_dim,
            hidden_dims=hidden_dims,
            critic_type=critic_type,
            **kwargs,
        ),
        neg_samples=neg_samples,
        gamma=gamma,
    )


def nwj(
    x_dim: int,
    y_dim: int,
    hidden_dims: List[int],
    neg_samples: int = 1,
    critic_type: str = JOINT_CRITIC,
    **kwargs,
) -> NWJ:
    return NWJ(
        critic=critic_nn(
            x_dim=x_dim,
            y_dim=y_dim,
            hidden_dims=hidden_dims,
            critic_type=critic_type,
            **kwargs,
        ),
        neg_samples=neg_samples,
    )


def smile(
    x_dim: int,
    y_dim: int,
    hidden_dims: List[int],
    neg_samples: int = 1,
    tau: float = 5.0,
    critic_type: str = JOINT_CRITIC,
    **kwargs,
) -> SMILE:
    return SMILE(
        critic=critic_nn(
            x_dim=x_dim,
            y_dim=y_dim,
            hidden_dims=hidden_dims,
            critic_type=critic_type,
            **kwargs,
        ),
        neg_samples=neg_samples,
        tau=tau,
    )


def tuba(
    x_dim: int,
    y_dim: int,
    hidden_dims: List[int],
    neg_samples: int = 1,
    critic_type: str = JOINT_CRITIC,
    **kwargs,
) -> TUBA:
    baseline_params = {}
    critic_params = {}

    for param_name, param_value in kwargs:
        if param_name in inspect.signature(baseline_nn).parameters:
            baseline_params[param_name] = param_value
        else:
            critic_params[param_name] = param_value

    b_nn = baseline_nn(x_dim=x_dim, hidden_dims=hidden_dims, **baseline_params)

    return TUBA(
        critic=critic_nn(
            x_dim=x_dim,
            y_dim=y_dim,
            hidden_dims=hidden_dims,
            critic_type=critic_type,
            **critic_params,
        ),
        baseline=b_nn,
        neg_samples=neg_samples,
    )



from typing import Callable, Any, Dict, Union, List, Tuple, Optional

import torch
from torch import nn

from torch_mist.estimators.base import MIEstimator
from torch_mist.utils.caching import cached_method
from torch_mist.utils.freeze import is_trainable
from torch_mist.estimators.transformed.utils import DummyModule

ERROR_MESSAGE = "The TransformedMIEstimator can be called by passing two arguments or by specifying multiple named named arguments."
SPLIT_SEQUENCE = "->"


class TransformedMIEstimator(MIEstimator):
    def __init__(
        self,
        base_estimator: MIEstimator,
        transforms: Optional[Dict[str, Callable[[Any], Any]]] = None,
        transforms_rename: Optional[
            Dict[Tuple[str, str], Callable[[Any], Any]]
        ] = None,
    ):
        super().__init__()

        assert transforms or transforms_rename
        if transforms_rename is None:
            transforms_rename = {}

        if transforms is None:
            transforms = {}

        self.infomax_gradient = base_estimator.infomax_gradient
        self.upper_bound = False
        self.lower_bound = base_estimator.lower_bound

        self.base_estimator = base_estimator
        self.transforms = nn.ModuleDict()

        additional_transforms = {
            (name, name): transform for name, transform in transforms.items()
        }
        transforms_rename.update(additional_transforms)
        for (
            variable_from,
            variable_to,
        ), transform in transforms_rename.items():
            if not base_estimator.infomax_gradient[variable_to] and (
                is_trainable(transform)
            ):
                print(
                    "Warning: Transforms can be trained together with the estimator only when the estimator provides a valid infomax gradient."
                    f"The estimator {base_estimator.__class__} does not. You can use a different estimator or the transform for {variable_from}."
                )

            if not isinstance(transform, nn.Module):
                transform = DummyModule(transform)

            if (
                SPLIT_SEQUENCE in variable_to
                or SPLIT_SEQUENCE in variable_from
            ):
                raise ValueError(
                    f"The estimators does not support variables that contain '{SPLIT_SEQUENCE}'."
                )

            self.transforms[
                f"{variable_from}{SPLIT_SEQUENCE}{variable_to}"
            ] = transform
            self.infomax_gradient[
                variable_from
            ] = base_estimator.infomax_gradient[variable_to]

    @cached_method
    def transform(self, **variables) -> Dict[str, torch.Tensor]:
        transformed_variables = {}
        for variable_fromto, transform in self.transforms.items():
            variable_from, variable_to = variable_fromto.split(SPLIT_SEQUENCE)
            transformed_variable = transform(variables[variable_from])

            # Detach the variables for which the gradient is not a valid infomax gradient
            if not self.infomax_gradient[variable_to]:
                transformed_variable = transformed_variable.detach()
            transformed_variables[variable_to] = transformed_variable

        return transformed_variables

    def _unfold_variables(self, *args, **variables) -> Dict[str, Any]:
        if len(args) != 0 and len(args) != 2:
            raise ValueError(ERROR_MESSAGE)
        if len(args) == 2:
            if len(variables) > 0:
                raise ValueError(ERROR_MESSAGE)
            variables["x"] = args[0]
            variables["y"] = args[1]

        return variables

    def batch_loss(self, *args, **variables) -> torch.Tensor:
        variables = self._unfold_variables(*args, **variables)
        variables.update(self.transform(**variables))
        return self.base_estimator.batch_loss(**variables)

    def log_ratio(
        self, *args, **variables
    ) -> Union[torch.Tensor, List[torch.Tensor]]:
        variables = self._unfold_variables(*args, **variables)
        variables.update(self.transform(**variables))
        return self.base_estimator.log_ratio(**variables)

    def unnormalized_log_ratio(self, *args, **variables) -> torch.Tensor:
        variables = self._unfold_variables(*args, **variables)
        variables.update(self.transform(**variables))
        return self.base_estimator.unnormalized_log_ratio(**variables)

    def loss(self, *args, **variables) -> torch.Tensor:
        return self.batch_loss(*args, **variables).mean()

    def mutual_information(
        self, *args, **variables
    ) -> Union[torch.Tensor, Dict[str, torch.Tensor]]:
        log_ratio = self.log_ratio(**variables)
        if isinstance(log_ratio, dict):
            return {k: v.mean() for k, v in log_ratio.items()}
        else:
            return log_ratio.mean()

    def forward(self, *args, **kwargs) -> torch.Tensor:
        return self.loss(*args, **kwargs)



import math
from typing import Dict

import torch

from pyro.distributions import ConditionalDistribution

from torch_mist.estimators.generative.base import (
    ConditionalGenerativeMIEstimator,
)
from torch_mist.utils.indexing import select_k_others


class L1Out(ConditionalGenerativeMIEstimator):
    infomax_gradient: Dict[str, bool] = {"x": True, "y": False}

    def __init__(
        self, q_Y_given_X: ConditionalDistribution, neg_samples: int = -1
    ):
        super().__init__(q_Y_given_X=q_Y_given_X)
        self.neg_samples = neg_samples

    def _broadcast_log_p_y_given_x(
        self, x: torch.Tensor, y: torch.Tensor
    ) -> torch.Tensor:
        assert (
            x.shape[0] == y.shape[0]
        ), "The batch size of x and y must be the same"
        N = x.shape[0]

        neg_samples = self.neg_samples
        if neg_samples <= 0:
            neg_samples = N + neg_samples

        if neg_samples < 1:
            neg_samples = 1

        if self.neg_samples == 0 or self.neg_samples == -1:
            x = x.unsqueeze(0)
        else:
            x = select_k_others(x, neg_samples)

        # Probability of all the other y in the same batch [M, N, ...]
        p_Y_given_x = self.q_Y_given_X.condition(x)
        log_p_y_given_x = p_Y_given_x.log_prob(y.unsqueeze(1) + x.detach() * 0)

        if self.neg_samples == -1:
            D = torch.eye(N).to(y.device)
            while D.ndim < log_p_y_given_x.ndim:
                D = D.unsqueeze(-1)

            # Remove the diagonal
            log_p_y_given_x = log_p_y_given_x * (1 - D)

            # Set the diagonal to -inf
            log_p_y_given_x = log_p_y_given_x + torch.nan_to_num(
                D * (-float("inf")),
                0,
                float("inf"),
                -float("inf"),
            )

        return log_p_y_given_x

    def approx_log_p_y(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        # log prob of the samples in the batch, shape [M, N, ...]
        log_p_y_given_x = self._broadcast_log_p_y_given_x(x, y)

        # Compute the expectation using logsumexp. The shape is [N, ...]
        log_p_y = torch.logsumexp(log_p_y_given_x, dim=0) - math.log(
            log_p_y_given_x.shape[0]
        )

        assert log_p_y.shape == y.shape[:-1]

        return log_p_y



import inspect
from typing import List, Union, Dict, Optional, Callable, Any

from pyro.distributions import ConditionalTransform, ConditionalDistribution
from pyro.nn import DenseNN

from torch.distributions import Transform, Distribution

from torch_mist.distributions.categorical import ConditionalCategoricalModule
from torch_mist.distributions.joint.base import JointDistribution
from torch_mist.distributions.normal import (
    ConditionalTransformedNormalModule,
    TransformedNormalModule,
    JointTransformedNormalModule,
)
from torch_mist.distributions.transforms import ConditionalDistributionModule

from torch_mist.distributions.transforms.utils import fetch_transform


def delete_unused_kwargs(
    transform_factory: Callable[[Any], Any],
    all_kwargs: Dict[str, Any],
    warnings: bool = True,
):
    # Check the arguments
    kwargs = {}
    for arg_name, value in all_kwargs.items():
        if arg_name in inspect.signature(transform_factory).parameters:
            kwargs[arg_name] = value
        elif warnings:
            name = transform_factory.__name__
            print(f"[Warning]: parameter {arg_name} ignored for {name}.")

    return kwargs


def make_transforms(
    input_dim: int,
    transform_name: str = "conditional_linear",
    normalization: Optional[str] = None,
    n_transforms: int = 1,
    **kwargs,
) -> List[Union[Transform, ConditionalTransform]]:
    assert n_transforms > 0, "n_transforms must be greater than 0"
    transforms = []

    kwargs["input_dim"] = input_dim

    transform_factory = fetch_transform(transform_name)
    transform_kwargs = delete_unused_kwargs(
        transform_factory, kwargs, warnings=True
    )

    if normalization:
        norm_factory = fetch_transform(normalization)
        norm_kwargs = delete_unused_kwargs(
            norm_factory, kwargs, warnings=False
        )
    else:
        norm_factory = None
        norm_kwargs = {}

    for transform in range(n_transforms):
        transform = transform_factory(**transform_kwargs)
        transforms.append(transform)
        if norm_factory:
            norm = norm_factory(**norm_kwargs)
            transforms.append(norm)

    return transforms


def transformed_normal(
    input_dim: int,
    transform_name: str = "linear",
    n_transforms: int = 1,
    normalization: Optional[str] = None,
    **kwargs,
) -> Distribution:
    assert n_transforms > 0, "n_transforms must be greater than 0"

    transforms = make_transforms(
        input_dim=input_dim,
        transform_name=transform_name,
        n_transforms=n_transforms,
        normalization=normalization,
        **kwargs,
    )

    return TransformedNormalModule(
        input_dim=input_dim,
        transforms=transforms,
    )


def conditional_transformed_normal(
    input_dim: int,
    context_dim: int,
    transform_name: str = "conditional_linear",
    n_transforms: int = 1,
    normalization: Optional[str] = None,
    **kwargs,
) -> ConditionalDistributionModule:
    assert n_transforms > 0, "n_transforms must be greater than 0"

    transforms = make_transforms(
        input_dim=input_dim,
        context_dim=context_dim,
        transform_name=transform_name,
        n_transforms=n_transforms,
        normalization=normalization,
        **kwargs,
    )

    return ConditionalTransformedNormalModule(
        input_dim=input_dim, transforms=transforms
    )


def joint_transformed_normal(
    input_dims: Dict[str, int],
    transform_name: str = "conditional_linear",
    n_transforms: int = 1,
    normalization: Optional[str] = None,
    **kwargs,
) -> JointDistribution:
    assert n_transforms > 0, "n_transforms must be greater than 0"

    input_dim = sum(input_dims.values())

    transforms = make_transforms(
        input_dim=input_dim,
        transform_name=transform_name,
        n_transforms=n_transforms,
        normalization=normalization,
        **kwargs,
    )

    return JointTransformedNormalModule(
        input_dims=input_dims, transforms=transforms
    )


def conditional_categorical(
    n_classes: int,
    context_dim: int,
    hidden_dims: List[int],
    temperature: float = 1.0,
):
    net = DenseNN(
        input_dim=context_dim, hidden_dims=hidden_dims, param_dims=[n_classes]
    )
    return ConditionalCategoricalModule(net, temperature=temperature)



import torch

from torch_mist.distributions.transforms.implementations.linear import Linear


class EMANormalize(Linear):
    def __init__(
        self,
        input_dim: int,
        epsilon: float = 1e-6,
        gamma: float = 0.99,
        normalize_inverse=True,
    ):
        super().__init__(input_dim=input_dim, epsilon=epsilon)

        assert 0 <= gamma <= 1
        self.gamma = gamma
        self.normalize_inverse = normalize_inverse

    def _update_params(self, t: torch.Tensor):
        self.loc = (
            self.gamma * self.loc + (1 - self.gamma) * t.mean(0).detach()
        )
        self.log_scale = (
            self.gamma * self.log_scale.exp()
            + (1 - self.gamma)
            * torch.clamp(t.std(0).detach(), min=self.epsilon)
        ).log()

    def _inverse(self, y: torch.Tensor) -> torch.Tensor:
        if self.training and self.normalize_inverse:
            self._update_params(y)
        return super()._inverse(y=y)

    def _call(self, x: torch.Tensor) -> torch.Tensor:
        if self.training and not self.normalize_inverse:
            self._update_params(x)
        return super()._call(x=x)



from typing import Optional, Union, Tuple, Dict, Callable, List, Any

import numpy as np
import pandas as pd
import torch
from torch.utils.data import DataLoader, random_split, Dataset, Subset

from torch_mist.estimators import MIEstimator, TransformedMIEstimator
from torch_mist.estimators.hybrid import PQHybridMIEstimator
from torch_mist.utils.data import SampleDataset, SameAttributeDataLoader
from torch_mist.utils.data.dataset import DataFrameDataset
from torch_mist.utils.data.loader import sample_same_value


TensorDictLike = Union[
    Tuple[Union[torch.Tensor, np.ndarray], Union[torch.Tensor, np.ndarray]],
    Dict[str, Union[torch.Tensor, np.ndarray]],
    Dataset,
    DataLoader,
    pd.DataFrame,
]


def prepare_variables(
    variables: Dict[str, torch.Tensor], device: torch.device
) -> Tuple[List[torch.Tensor], Dict[str, torch.Tensor]]:
    v_args, v_kwargs = [], {}
    if torch.is_tensor(variables):
        variables = [variables]
    if isinstance(variables, dict):
        v_kwargs = {k: v.to(device) for k, v in variables.items()}
    elif isinstance(variables, tuple) or isinstance(variables, list):
        v_args = [v.to(device) for v in variables]

    return v_args, v_kwargs


def infer_dims(
    data: TensorDictLike,
) -> Dict[str, int]:
    dataloader, _ = make_default_dataloaders(
        data=data,
        valid_data=None,
        batch_size=1,
        valid_percentage=0,
        num_workers=0,
    )

    batch = next(iter(dataloader))
    if not isinstance(batch, dict):
        assert len(batch) == 2
        batch = {"x": batch[0], "y": batch[1]}

    return {k: v.shape[-1] for k, v in batch.items()}


def convert_to_tensor(
    array: Union[torch.Tensor, np.array]
) -> torch.FloatTensor:
    if isinstance(array, np.ndarray):
        array = torch.FloatTensor(array)
    return array


def is_valid_entry(
    entry: Union[Dict[str, Any], np.ndarray, torch.Tensor, Tuple[Any, ...]]
) -> bool:
    if isinstance(entry, torch.Tensor):
        return torch.sum(entry != entry) == 0
    elif isinstance(entry, np.ndarray):
        return np.sum(entry != entry) == 0
    elif hasattr(entry, "values"):
        is_valid = True
        for element in entry.values():
            if not is_valid_entry(element):
                is_valid = False
                break
        return is_valid
    elif isinstance(entry, tuple):
        is_valid = True
        for element in entry:
            if not is_valid_entry(element):
                is_valid = False
                break
        return is_valid
    else:
        return True


def make_dataset(data: TensorDictLike) -> Dataset:
    # Validate the input combinations
    if isinstance(data, list) or isinstance(data, tuple):
        if len(data) != 2:
            raise ValueError("data should be a tuple of two elements (x, y).")
        x, y = data[0], data[1]

        x = convert_to_tensor(x)
        y = convert_to_tensor(y)

        # Unsqueeze an empty dimension if necesary
        if len(x.shape) == 1:
            x = x.view(-1, 1)
        if len(y.shape) == 1:
            y = y.view(-1, 1)

        dataset = SampleDataset({"x": x, "y": y})
    elif isinstance(data, dict):
        dataset = SampleDataset(data)
    elif isinstance(data, pd.DataFrame):
        dataset = DataFrameDataset(data)
    else:
        dataset = data

    return dataset


def filter_dataset(dataset: Dataset):
    # Remove invalid entries
    valid_ids = []
    for idx, entry in enumerate(dataset):
        if is_valid_entry(entry):
            valid_ids.append(idx)
    if len(valid_ids) != len(dataset):
        print(
            f"[Warning]: Removing {len(dataset)-len(valid_ids)} entries from the dataset"
        )
        dataset = Subset(dataset, valid_ids)

    return dataset


def is_data_loader(data: TensorDictLike):
    return isinstance(data, DataLoader)


def make_default_dataloader(
    data: TensorDictLike,
    batch_size: Optional[int] = None,
    num_workers: int = 0,
    filter_invalid_data: bool = True,
) -> DataLoader:
    if is_data_loader(data):
        dataloader = data
    else:
        if batch_size is None:
            raise ValueError("Please specify a value for batch_size.")
        dataset = make_dataset(data)

        if filter_invalid_data:
            dataset = filter_dataset(dataset)

        dataloader = DataLoader(
            dataset,
            batch_size=batch_size,
            num_workers=num_workers,
            shuffle=True,
        )
    return dataloader


def make_default_dataloaders(
    data: TensorDictLike,
    valid_data: Optional[TensorDictLike] = None,
    valid_percentage: float = 0.1,
    batch_size: Optional[int] = None,
    filter_invalid_data: bool = True,
    num_workers: int = 0,
) -> Tuple[DataLoader, Optional[DataLoader]]:
    # Make the datasets if necessary
    if is_data_loader(data):
        train_loader = data
        train_set = data.dataset
        batch_size = data.batch_size
        num_workers = data.num_workers
    else:
        if batch_size is None:
            raise ValueError("Please specify a value for batch_size.")
        train_set = make_dataset(data)
        train_loader = None

    # Filter out NaNs
    if filter_invalid_data:
        train_set = filter_dataset(train_set)

    if valid_data is None:
        valid_set = None
        valid_loader = None
    elif is_data_loader(valid_data):
        valid_loader = valid_data
        valid_set = None
    else:
        valid_set = make_dataset(valid_data)
        valid_loader = None

    if valid_set and filter_invalid_data:
        valid_set = filter_dataset(valid_set)

    # Create a validation set if specified
    if valid_percentage > 0:
        if valid_set is None and valid_loader is None:
            # Make a random train/valid split
            n_valid = int(len(train_set) * valid_percentage)
            train_set, valid_set = random_split(
                train_set, [len(train_set) - n_valid, n_valid]
            )
            train_loader = None
        else:
            print(
                "[Warning]: valid_percentage is ignored since valid_set or valid_loader are already specified."
            )

    if train_loader is None:
        train_loader = DataLoader(
            train_set,
            batch_size=batch_size,
            shuffle=True,
            num_workers=num_workers,
        )

    if valid_loader is None and not (valid_set is None):
        valid_loader = DataLoader(
            valid_set,
            batch_size=batch_size,
            num_workers=num_workers,
            shuffle=True,
        )

    assert is_data_loader(train_loader)
    assert is_data_loader(valid_loader) or (valid_loader is None)

    return train_loader, valid_loader


def update_dataloader(
    estimator: MIEstimator,
    dataloader: DataLoader,
) -> DataLoader:
    # TODO: this is kinda hacky.. find a better way

    # Check if we need to modify the dataloader to sample the same attributes
    _estimator = estimator
    transforms = []
    while isinstance(_estimator, TransformedMIEstimator):
        transforms.append(_estimator.transforms)
        _estimator = estimator.base_estimator

    # If required, change the data-loader to sample batches with the same attribute only
    if isinstance(_estimator, PQHybridMIEstimator):
        neg_samples = _estimator.neg_samples
        device = next(iter(estimator.parameters())).device
        transforms.append({"y->y": _estimator.quantize_y})

        def compute_attributes(samples) -> torch.Tensor:
            v_args, v_kwargs = prepare_variables(samples, device)

            if "y" in v_kwargs:
                y = v_kwargs["y"]
            else:
                assert len(v_args) == 2
                y = v_args[1]

            for transform in transforms:
                y = transform["y->y"](y)
            return y.data.cpu()

        if not isinstance(dataloader, SameAttributeDataLoader):
            dataloader = sample_same_value(
                dataloader, compute_attributes, neg_samples=neg_samples
            )

    return dataloader



from typing import Callable, Optional

from pyro.distributions import Delta

from torch_mist.distributions.transforms import ConditionalDistributionModule


class ConditionalDelta(ConditionalDistributionModule):
    def __init__(self, transform: Optional[Callable] = None):
        super().__init__()
        self.transform = transform

    def condition(self, context):
        if not (self.transform is None):
            context = self.transform(context)
        return Delta(context)



from .base import MIEstimator
from .generative.implementations import *
from .discriminative.implementations import *
from .transformed.implementations import *
from .transformed import TransformedMIEstimator
from .hybrid import *
from .factories import *
from .utils import FlippedMIEstimator, flip_estimator
from .multi import MultiMIEstimator



import inspect
from abc import abstractmethod
from typing import Optional, Type, Dict, List, Union

import torch
from pyro.distributions import Delta
from torch.distributions import (
    Distribution,
    Normal,
    Laplace,
    Categorical,
    OneHotCategorical,
)
from torch.nn.functional import softplus

EPSILON = 1e-6


class ParameterMap:
    """
    Mapping from a list of tensors to a dictionary representing the parametrization of a distribution.
    Abstract class.
    """

    supported_distributions: List[Type[Distribution]] = []
    n_parameters: int = 0

    @abstractmethod
    def map_parameters(self, parameter_list: List[torch.Tensor]) -> Dict[str, torch.Tensor]:
        """Abstract method responsible for defining the mapping from the list of tensors to the parameter dictionary.

        Args:
            parameter_list (List[torch.Tensor]): The list of parameters for the distribution of interest.

        Returns:
            Dict[str, torch.Tensor]: The named parameters for the distribution of interest.
        """
        pass

    def __call__(
        self, parameter_list: Union[torch.Tensor, List[torch.Tensor]]
    ) -> Dict[str, torch.Tensor]:
        if isinstance(parameter_list, torch.Tensor):
            parameter_list = [parameter_list]

        assert len(parameter_list) == self.n_parameters

        mapped_parameters = self.map_parameters(parameter_list)

        # Check the mapping is compatible with the specified distributions
        for distribution_class in self.supported_distributions:
            for parameter_name in mapped_parameters:
                assert parameter_name in inspect.signature(distribution_class).parameters

        return mapped_parameters


class DeltaMap(ParameterMap):
    """
    Delta Parameter mapping: maps a list of one tensor into a delta distribution at the same location
    """

    supported_distributions: List[Type[Distribution]] = [Delta]

    n_parameters: int = 1

    def map_parameters(self, parameter_list: List[torch.Tensor]) -> Dict[str, torch.Tensor]:
        return {"v": parameter_list[0]}


class LocScaleMap(ParameterMap):
    supported_distributions: List[Type[Distribution]] = [Normal, Laplace]

    def __init__(self, scale: Optional[float] = None, epsilon: float = EPSILON, scale_offset: float = 0.0):
        """Location and scale parametrization that can be used for Normal and Laplace distributions. The scale is constrained to be positive. This is obtained by applying a softplus function.

        Args:
            scale (Optional[float], optional): The fixed scale of the distribution. If passed, only the tensor list maps to the scale only. Defaults to None.
            epsilon (float, optional): Small float used to increased the numerical stability. Defaults to EPSILON.
        """
        if scale:
            n_parameters = 1
        else:
            n_parameters = 2

        self.scale = scale
        self.epsilon = epsilon
        self.n_parameters = n_parameters
        self.scale_offset = scale_offset

    def map_parameters(self, parameter_list: List[torch.Tensor]) -> Dict[str, torch.Tensor]:
        loc = parameter_list[0]
        if self.n_parameters == 1:
            scale = self.scale
        else:
            log_scale = parameter_list[1]
            # Numerically stable scale parametrization
            scale = softplus(log_scale+self.scale_offset) + self.epsilon
            assert torch.all(scale > 0), f"Scale must be positive, got {scale}"

        return {"loc": loc, "scale": scale}


class LogitsMap(ParameterMap):
    supported_distributions: List[Type[Distribution]] = [
        Categorical,
        OneHotCategorical,
    ]
    n_parameters: int = 1

    def map_parameters(self, parameter_list: List[torch.Tensor]) -> Dict[str, torch.Tensor]:
        logits = parameter_list[0]
        return {"logits": logits}



import torch
from torch.distributions import Distribution

from torch_mist.utils.indexing import select_k_others


class EmpiricalDistribution(Distribution):
    def __init__(self):
        super().__init__(validate_args=False)
        self._samples = None

    def add_samples(self, samples):
        self._samples = samples

    def sample(self, sample_shape: torch.Size = torch.Size()) -> torch.Tensor:
        assert len(sample_shape) == 1
        n_samples = sample_shape[0]
        return select_k_others(self._samples, n_samples)

    def update(self):
        self._samples = None

    def __repr__(self):
        return f"{self.__class__.__name__}(n_samples={None if self._samples is None else len(self._samples)})"



from contextlib import contextmanager

import torch

from torch_mist.baseline import BatchLogMeanExp, ExponentialMovingAverage
from torch_mist.critic import Critic
from torch_mist.estimators.discriminative.base import (
    BaselineDiscriminativeMIEstimator,
)


class MINE(BaselineDiscriminativeMIEstimator):
    lower_bound = False  # Technically MINE is a lower bound, but sometimes it converges from above

    def __init__(
        self,
        critic: Critic,
        neg_samples: int = 1,
        gamma: float = 0.9,
    ):
        super().__init__(
            critic=critic,
            baseline=BatchLogMeanExp("all"),
            neg_samples=neg_samples,
        )
        self._train_baseline = ExponentialMovingAverage(gamma=gamma)

    @contextmanager
    def train_baseline(self):
        eval_baseline = self.baseline
        self.baseline = self._train_baseline
        try:
            yield
        finally:
            # Restore the original baseline
            self.baseline = eval_baseline

    def batch_loss(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        with self.train_baseline():
            batch_loss = super().batch_loss(x, y)

        return batch_loss



from abc import abstractmethod
from typing import Dict, Tuple, Optional

import torch

from torch_mist.baseline import Baseline
from torch_mist.estimators.base import MIEstimator
from torch_mist.critic import Critic
from torch_mist.critic import SeparableCritic
from torch_mist.distributions.empirical import EmpiricalDistribution
from torch_mist.utils.caching import cached_method
from torch_mist.utils.indexing import matrix_off_diagonal


class DiscriminativeMIEstimator(MIEstimator):
    lower_bound: bool = True
    infomax_gradient: Dict[str, bool] = {"x": True, "y": True}

    def __init__(
        self,
        critic: Critic,
        neg_samples: int = 1,
    ):
        super().__init__()
        self.critic = critic
        self.neg_samples = neg_samples
        self.proposal = EmpiricalDistribution()

    @cached_method
    def unnormalized_log_ratio(
        self, x: torch.Tensor, y: torch.Tensor
    ) -> torch.Tensor:
        f = self.critic(x, y)
        assert f.ndim == y.ndim - 1
        return f

    def n_negatives_to_use(self, N: int):
        neg_samples = self.neg_samples

        # Negative neg_samples values are interpreted as difference from the batch size (-1 is all but one in the batch)
        if neg_samples <= 0:
            neg_samples = N + neg_samples

        # We can't use more negative than the batch
        if neg_samples > N:
            neg_samples = N

        # At least one negative sample
        neg_samples = max(neg_samples, 1)
        return neg_samples

    @cached_method
    def sample_negatives(
        self, x: torch.Tensor, y: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, Optional[torch.Tensor]]:
        N = x.shape[0]
        neg_samples = self.n_negatives_to_use(N)

        if isinstance(self.proposal, EmpiricalDistribution):
            self.proposal.add_samples(y)

        # Efficient implementation for separable critic with empirical distribution (negatives from the same batch)
        if (
            isinstance(self.critic, SeparableCritic)
            and self.neg_samples <= 0
            and isinstance(self.proposal, EmpiricalDistribution)
        ):
            y_ = self.proposal._samples[:N].unsqueeze(1)
        else:
            # Sample from the proposal p(y) [M, ..., Y_DIM] with M as the number of neg_samples
            y_ = self.proposal.sample(torch.Size([neg_samples]))
            # The shape of the samples from the proposal distribution is [M, ..., Y_DIM]
            assert y_.ndim == x.ndim + 1 and y_.shape[0] == neg_samples
            assert y_.shape[0] == neg_samples and y_.ndim == x.ndim + 1

        if isinstance(self.proposal, EmpiricalDistribution):
            self.proposal.update()

        return x.unsqueeze(0), y_, None

    def log_ratio(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        # Approximate the log-ratio p(x,y)/r(x,y) on samples from p(x,y).
        # x and y have shape [..., X_DIM] and [..., Y_DIM] respectively

        assert x.ndim == y.ndim
        # Sample x_ and y_ from r(x,y). w represents an optional log-weighting coefficient log[r(x,y)/q(x,y)] in case
        # x_ and y_ are sampled from q(x,y) instead of r(x,y).
        # By default we have r(x,y) = p(x)p(y) and shape [M, ..., X_DIM] and [M, ..., Y_DIM] respectively
        x_, y_, log_w = self.sample_negatives(x, y)

        # Evaluate the unnormalized_log_ratio f(x,y) on the samples from p(x, y), with shape [...]
        unnormalized_log_ratio = self.unnormalized_log_ratio(x, y)

        x_dim = x_.ndim + (1 if isinstance(x_, torch.LongTensor) else 0)
        y_dim = y_.ndim + (1 if isinstance(y_, torch.LongTensor) else 0)

        assert x_dim == y_dim

        # Compute the log-normalization term log E{r(x,y)}[e^f(x,y)] on samples from r(x,y)
        log_partition = self.approx_log_partition(
            x=x, y=y, x_=x_, y_=y_, log_w=log_w
        )

        assert log_partition.shape == unnormalized_log_ratio.shape
        log_ratio = unnormalized_log_ratio - log_partition

        return log_ratio

    @abstractmethod
    def _approx_log_partition(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        f_: torch.Tensor,
        log_w: Optional[torch.Tensor],
    ) -> torch.Tensor:
        raise NotImplementedError()

    def approx_log_partition(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        x_: torch.Tensor,
        y_: torch.Tensor,
        log_w: Optional[torch.Tensor],
    ) -> torch.Tensor:
        N = x.shape[0]
        M = self.n_negatives_to_use(N)

        # Evaluate the unnormalized_log_ratio f(x_,y_) on the samples x_, y_ ~ r(x, y). It has shape [M, ...]
        f_ = self.critic(x_, y_)

        # Computational shortcut for separable critic
        # we compute the product for the whole batch and then remove the diagonal
        if self.neg_samples < 0 and M != f_.shape[0]:
            assert isinstance(self.critic, SeparableCritic)
            # Remove the diagonal from the matrix
            f_ = matrix_off_diagonal(f_, M)

        assert f_.shape[0] == M and f_.shape[1] == N

        log_Z = self._approx_log_partition(x=x, y=y, f_=f_, log_w=log_w)
        assert log_Z.shape == x.shape[:-1]

        return log_Z

    def batch_loss(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return -self.log_ratio(x, y)

    def __repr__(self):
        s = self.__class__.__name__ + "(\n"
        s += (
            "  (critic): " + str(self.critic).replace("\n", "\n" + "  ") + "\n"
        )
        s += "  (neg_samples): " + str(self.neg_samples) + "\n"
        s += ")"

        return s


class BaselineDiscriminativeMIEstimator(DiscriminativeMIEstimator):
    def __init__(
        self,
        critic: Critic,
        baseline: Baseline,
        neg_samples: int = 1,
    ):
        super().__init__(
            critic=critic,
            neg_samples=neg_samples,
        )

        self.baseline = baseline

    def _approx_log_partition(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        f_: torch.Tensor,
        log_w: Optional[torch.Tensor],
    ) -> torch.Tensor:
        # Compute the baseline
        b = self.baseline(x=x, f_=f_).unsqueeze(0)
        assert (
            b.ndim == f_.ndim
        ), f"Baseline has ndim {b.ndim} while f_ has ndim {f_.ndim}"

        # Add the log_weights if provided
        if not (log_w is None):
            assert log_w.ndim == f_.ndim
            f_ = f_ + log_w

        # Compute the log_partition, it has shape [...]
        log_Z = (f_ - b).exp() + b - 1.0
        return log_Z.mean(0)

    def __repr__(self):
        s = self.__class__.__name__ + "(\n"
        s += (
            "  (ratio_estimator): "
            + str(self.critic).replace("\n", "\n" + "  ")
            + "\n"
        )
        s += (
            "  (baseline): "
            + str(self.baseline).replace("\n", "\n" + "  ")
            + "\n"
        )
        s += "  (neg_samples): " + str(self.neg_samples) + "\n"
        s += ")"

        return s



from abc import abstractmethod
from typing import List, Callable, Optional, Tuple, Any

import torch
import torch.nn as nn
from pyro.nn import DenseNN

from torch_mist.utils.caching import cached_method


class Model(nn.Module):
    lower_bound: bool = False
    upper_bound: bool = False

    @abstractmethod
    def loss(self, *args, **kwargs):
        raise NotImplementedError()


class Identity(nn.Module):
    def forward(self, x: Any):
        return x


class Normalize(nn.Module):
    def forward(self, x):
        return x / (x**2).sum(-1).unsqueeze(-1) ** 0.5


class CachedModule(nn.Module):
    def __init__(self, module: nn.Module):
        super().__init__()
        self.module = module

    @cached_method
    def forward(self, *args, **kwargs):
        return self.module.forward(*args, **kwargs)

    def __repr__(self):
        return "cached_" + self.module.__repr__()


def dense_nn(
    input_dim: int,
    output_dim: int,
    hidden_dims: Optional[List[int]] = None,
    nonlinearity: Optional[Callable[[torch.Tensor], torch.Tensor]] = None,
) -> nn.Module:
    if nonlinearity is None:
        nonlinearity = nn.ReLU(True)

    if hidden_dims is None:
        hidden_dims = []
    if len(hidden_dims) == 0:
        return nn.Linear(input_dim, output_dim)
    else:
        return DenseNN(
            input_dim=input_dim,
            hidden_dims=hidden_dims,
            param_dims=[output_dim],
            nonlinearity=nonlinearity,
        )


def multi_head_dense_nn(
    input_dim: int,
    output_dim: int,
    n_shared_layers: int,
    hidden_dims: List[int],
    n_heads: int,
    cached_shared_forward: bool = True,
    nonlinearity: nn.Module = nn.ReLU(True),
) -> Tuple[nn.Module, ...]:
    hidden_dims = [input_dim] + hidden_dims

    if n_shared_layers < 0:
        n_shared_layers = len(hidden_dims) + n_shared_layers

    assert 0 <= n_shared_layers <= len(hidden_dims)

    if n_shared_layers == 0:
        shared_net = None
    else:
        shared_net = dense_nn(
            input_dim=input_dim,
            output_dim=hidden_dims[n_shared_layers],
            hidden_dims=hidden_dims[1:n_shared_layers],
            nonlinearity=nonlinearity,
        )
        shared_net = nn.Sequential(
            shared_net,
            nonlinearity,
        )

        if cached_shared_forward:
            shared_net = CachedModule(shared_net)

    nets = []
    for _ in range(n_heads):
        head = dense_nn(
            input_dim=hidden_dims[n_shared_layers],
            output_dim=output_dim,
            hidden_dims=hidden_dims[n_shared_layers + 1 :],
            nonlinearity=nonlinearity,
        )
        if n_shared_layers == 0:
            nets.append(head)
        elif n_shared_layers == len(hidden_dims):
            nets.append(shared_net)
        else:
            nets.append(nn.Sequential(shared_net, head))

    return tuple(nets)



from typing import Callable, Any

from torch import nn


class DummyModule(nn.Module):
    def __init__(self, f: Callable[[Any], Any]):
        super().__init__()
        self.f = f

    def forward(self, *args, **kwargs):
        return self.f(*args, **kwargs)

    def __repr__(self):
        return str(self.f)



def fetch_transform(transform_name: str):
    import pyro.distributions.transforms as pyro_transforms_module
    import torch_mist.distributions.transforms.factories as transforms_module

    if hasattr(pyro_transforms_module, transform_name):
        transform_factory = getattr(pyro_transforms_module, transform_name)
    elif hasattr(transforms_module, transform_name):
        transform_factory = getattr(transforms_module, transform_name)
    else:
        raise NotImplementedError(
            f"Transform {transform_name} is not implemented."
        )
    return transform_factory



from typing import Optional, List, Union

from pyro.distributions import ConditionalDistribution

from torch_mist.estimators.transformed.implementations import (
    PQ,
    BinnedMIEstimator,
)
from torch_mist.quantization import (
    QuantizationFunction,
    instantiate_quantization,
)


def binned(
    quantize_x: Optional[Union[QuantizationFunction, str]] = "kmeans",
    quantize_y: Optional[Union[QuantizationFunction, str]] = "kmeans",
    temperature: float = 0.1,
    n_bins: Optional[int] = 32,
    **kwargs
) -> BinnedMIEstimator:
    if isinstance(quantize_x, str):
        quantize_x = instantiate_quantization(
            quantize_x,
            n_bins=n_bins,
            **kwargs,
        )
    if isinstance(quantize_y, str):
        quantize_y = instantiate_quantization(
            quantize_y, n_bins=n_bins, **kwargs
        )

    return BinnedMIEstimator(
        quantize_x=quantize_x,
        quantize_y=quantize_y,
        temperature=temperature,
    )


def pq(
    quantize_y: Optional[Union[QuantizationFunction, str]] = "kmeans",
    x_dim: Optional[int] = None,
    hidden_dims: Optional[List[int]] = None,
    q_QY_given_X: Optional[ConditionalDistribution] = None,
    temperature: float = 0.1,
    n_bins: Optional[int] = 32,
    **kwargs
) -> PQ:
    from torch_mist.distributions.factories import conditional_categorical

    if isinstance(quantize_y, str):
        quantize_y = instantiate_quantization(
            quantize_y,
            n_bins=n_bins,
            **kwargs,
        )

    if q_QY_given_X is None:
        if x_dim is None or hidden_dims is None:
            raise ValueError(
                "Either q_QY_given_X or x_dim and hidden_dims must be specified."
            )
        q_QY_given_X = conditional_categorical(
            n_classes=quantize_y.n_bins,
            context_dim=x_dim,
            hidden_dims=hidden_dims,
            temperature=temperature,
        )

    return PQ(
        q_QY_given_X=q_QY_given_X,
        quantize_y=quantize_y,
        temperature=temperature,
    )



import torch

from .base import Critic
from torch_mist.utils.shape import expand_to_same_shape


class ConstantCritic(Critic):
    def __init__(self, value: float = 0):
        super().__init__()
        self.value = value

    def forward(self, x, y) -> torch.Tensor:
        x, y = expand_to_same_shape(x, y)
        return x[..., 0].detach() * 0 + self.value



import collections
from typing import Optional, Union, Dict

import torch
import numpy as np
from torch import nn
from torch.utils.data import DataLoader

from torch_mist.utils.data.utils import prepare_variables, TensorDictLike


def evaluate(
    model: nn.Module,
    method: str,
    data: TensorDictLike,
    device: torch.device = torch.device("cpu"),
    batch_size: Optional[int] = None,
    num_workers: int = 0,
) -> Union[float, Dict[str, float]]:
    if not hasattr(model, method):
        raise ValueError(f"{model.__class__.__name__} has no method {method}.")

    # Make a dataloader if necessary
    if isinstance(data, DataLoader):
        dataloader = data
    else:
        if batch_size is None:
            raise ValueError("Plase specify a value for batch_size.")
        dataloader = DataLoader(
            data,
            batch_size=batch_size,
            num_workers=num_workers,
        )

    model.eval()
    model = model.to(device)

    values = None
    for samples in dataloader:
        v_args, v_kwargs = prepare_variables(samples, device)
        estimation = getattr(model, method)(*v_args, **v_kwargs)

        if values is None:
            if isinstance(estimation, dict):
                values = collections.defaultdict(list)
            else:
                values = []

        if isinstance(estimation, dict):
            for key, value in estimation.items():
                values[key].append(value.item())
        else:
            values.append(estimation.item())

    if isinstance(values, dict):
        for k, v in values.items():
            values[k] = np.mean(v)
    else:
        values = np.mean(values)

    return values



from abc import abstractmethod
from typing import Optional, Any

import numpy as np
from sklearn.base import TransformerMixin
import torch
from torch import nn
from torch.utils.data import DataLoader

from torch_mist.utils.freeze import freeze


class QuantizationFunction(nn.Module):
    @property
    def n_bins(self) -> int:
        raise NotImplementedError()

    @abstractmethod
    def quantize(self, x: torch.Tensor) -> torch.LongTensor:
        raise NotImplementedError()

    def forward(self, x: torch.Tensor) -> torch.LongTensor:
        return self.quantize(x)


class NotTrainedError(Exception):
    def __init__(self, message: str, model_to_train: nn.Module):
        super().__init__(message)
        self.model_to_train = model_to_train


class LearnableQuantization(QuantizationFunction):
    def __init__(self, **train_params):
        super().__init__()
        self.trained = False
        self.train_params = train_params

    def _fit(self, dataloader: DataLoader) -> Optional[Any]:
        from torch_mist.utils.train.model import train_model

        return train_model(
            model=self, train_data=dataloader, **self.train_params
        )

    def fit(self, dataloader: Any) -> Optional[Any]:
        log = self._fit(dataloader)
        self.trained = True
        freeze(self)
        return log

    def forward(self, x: torch.Tensor) -> torch.LongTensor:
        if not self.trained:
            raise NotTrainedError(
                f"The {self.__class__.__name__} quantization scheme needs to be trained.\n"
                + "Please call the fit(data, **kwargs) method before quantizing.",
                self,
            )
        return super().forward(x)

    def loss(self, *args, **kwargs) -> torch.Tensor:
        raise NotImplementedError()


class ClusterQuantization(LearnableQuantization):
    def __init__(self, clustering: TransformerMixin, **train_params):
        super().__init__(**train_params)
        self.clustering = clustering

    @property
    def n_bins(self) -> int:
        return self.clustering.n_clusters

    def _fit(self, dataloader: DataLoader):
        data = []
        for batch in dataloader:
            assert isinstance(batch, torch.Tensor)
            data.append(batch.cpu().data.numpy())

        data = np.concatenate(data, 0).astype(np.float32)

        if not isinstance(data, np.ndarray):
            raise ValueError(
                "data needs to be an instance of torch.Tensor or np.array."
            )

        self.clustering.fit(data, **self.train_params)

    def quantize(self, x: torch.Tensor) -> torch.LongTensor:
        device = x.device
        shape = x.shape[:-1]
        feature_dim = x.shape[-1]
        x = x.view(-1, feature_dim)

        if isinstance(x, torch.Tensor):
            x = x.cpu().data.numpy()

        return (
            torch.LongTensor(self.clustering.predict(x)).view(shape).to(device)
        )


class VectorQuantization(QuantizationFunction):
    def __init__(
        self,
        vectors: torch.Tensor,
    ):
        super().__init__()
        self.register_buffer("vectors", vectors)

    def codebook_lookup(self, x: torch.Tensor) -> torch.Tensor:
        assert x.shape[-1] == self.vectors.shape[-1]
        shape = x.shape[:-1]

        x = x.reshape(-1, 1, x.shape[-1])

        # Compute distances between z and vectors
        distances = torch.cdist(x, self.vectors.unsqueeze(0))
        # Get the closest vector
        _, indices = torch.min(distances, dim=-1)
        # Get the corresponding vector

        return indices.reshape(shape)

    @property
    def n_bins(self) -> int:
        return self.vectors.shape[0]

    def quantize(self, x: torch.Tensor) -> torch.LongTensor:
        indices = self.codebook_lookup(x)
        return indices.long()


class LearnableVectorQuantization(VectorQuantization, LearnableQuantization):
    def __init__(
        self,
        transform: Optional[nn.Module] = None,
        vectors: Optional[torch.Tensor] = None,
        n_bins: Optional[int] = None,
        quantization_dim: Optional[int] = None,
        **train_params,
    ):
        if vectors is None:
            if n_bins is None or quantization_dim is None:
                raise ValueError(
                    "Either `vectors` or `n_bins` and `input_dim` need to be specified"
                )

            # Vectors used for quantization
            vectors = torch.zeros(n_bins, quantization_dim)
            vectors.uniform_(-1.0 / n_bins, 1.0 / n_bins)
        else:
            if vectors.ndim != 2:
                raise NotImplementedError(
                    "Vector quantization is currently supported only on one dimension"
                )
            if quantization_dim:
                if quantization_dim != vectors.shape[-1]:
                    raise ValueError(
                        "The specified quantization_dim is not compatible with the vectors."
                    )
            if n_bins:
                if n_bins != vectors.shape[0]:
                    raise ValueError(
                        "The specified n_bins differs from the number of quantization vectors"
                    )

        VectorQuantization.__init__(self, vectors=vectors)
        self.train_params = train_params
        self.trained = False
        self.transform = transform

    def quantize(self, x: torch.Tensor) -> torch.LongTensor:
        if self.transform:
            x = self.transform(x)
        return super().quantize(x)


class FixedQuantization(QuantizationFunction):
    def __init__(self, input_dim: int, thresholds: torch.Tensor):
        super().__init__()
        self.thresholds = thresholds
        self.input_dim = input_dim

    @property
    def n_bins(self) -> int:
        return (self.thresholds.shape[0] + 1) ** self.input_dim

    def quantize(self, x: torch.Tensor) -> torch.LongTensor:
        bins = torch.bucketize(x, self.thresholds)
        flat_bins = (
            bins
            * (
                torch.FloatTensor([self.thresholds.shape[0] + 1])
                ** torch.arange(self.input_dim)
            ).to(bins.device)
        ).sum(-1)
        return flat_bins.long()



from typing import Dict

import torch


from torch_mist.distributions.conditional.unconditional import (
    UnconditionalDistribution,
)
from torch_mist.distributions.empirical import EmpiricalDistribution
from torch_mist.estimators.generative.base import (
    ConditionalGenerativeMIEstimator,
)
from torch_mist.utils.shape import expand_to_same_shape


class DummyGenerativeMIEstimator(ConditionalGenerativeMIEstimator):
    lower_bound: bool = True
    infomax_gradient: Dict[str, bool] = {"x": True, "y": False}

    def __init__(
        self,
    ):
        super().__init__(
            q_Y_given_X=UnconditionalDistribution(EmpiricalDistribution())
        )

    def log_ratio(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        x, y = expand_to_same_shape(x, y)
        return x[..., 0] * 0

    def batch_loss(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return -self.log_ratio(x, y)



from typing import Optional, Union
from functools import lru_cache

import torch
from torch.distributions import Distribution

from torch_mist.distributions.joint.base import JointDistribution
from torch_mist.estimators.generative.base import JointGenerativeMIEstimator
from torch_mist.utils.freeze import is_trainable


class GM(JointGenerativeMIEstimator):
    def __init__(
        self,
        q_XY: JointDistribution,
        q_Y: Union[Distribution, JointDistribution],
        q_X: Union[Distribution, JointDistribution],
    ):
        super().__init__(q_XY=q_XY)
        self._q_Y = q_Y
        self._q_X = q_X

    @property
    def q_X(self):
        return self._q_X

    @property
    def q_Y(self):
        return self._q_Y

    def batch_loss(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
    ) -> torch.Tensor:
        batch_loss = super().batch_loss(x=x, y=y)

        if is_trainable(self.q_Y):
            log_q_y = self.approx_log_p_y(x=x, y=y)
            batch_loss = batch_loss - log_q_y

        if is_trainable(self.q_X):
            log_q_x = self.approx_log_p_x(x=x, y=y)
            batch_loss = batch_loss - log_q_x

        assert (
            batch_loss.shape == y.shape[:-1] and torch.is_floating_point(y)
        ) or (batch_loss.shape == y.shape and not torch.is_floating_point(y))

        return batch_loss



from typing import Type, Optional, Dict, Any, Union, Tuple, List, Callable

import torch
from torch.optim import Optimizer
from torch.optim import Adam
from torch.utils.data import DataLoader

from torch_mist.estimators.base import MIEstimator
from torch_mist.utils.data.dataset import WrappedDataset

from torch_mist.utils.logging.logger.base import Logger
from torch_mist.utils.data.utils import (
    TensorDictLike,
    make_default_dataloaders,
    update_dataloader,
)
from torch_mist.utils.train.model import train_model


def pretrain_components(
    estimator: MIEstimator, train_loader: DataLoader, verbose: bool = True
):
    for func, component in estimator._components_to_pretrain:
        trained = False
        if hasattr(component, "trained"):
            trained = component.trained
        if not trained:
            wrapped_loader = DataLoader(
                WrappedDataset(train_loader.dataset, func),
                batch_size=train_loader.batch_size,
                shuffle=True,
                num_workers=train_loader.num_workers,
            )
            if verbose:
                print(f"Training {component.__class__.__name__}()")
            component.fit(wrapped_loader)


def train_mi_estimator(
    estimator: MIEstimator,
    train_data: TensorDictLike,
    valid_data: Optional[TensorDictLike] = None,
    valid_percentage: float = 0.1,
    batch_size: Optional[int] = None,
    num_workers: int = 0,
    device: Union[torch.device, str] = torch.device("cpu"),
    max_epochs: Optional[int] = None,
    max_iterations: Optional[int] = None,
    optimizer_class: Type[Optimizer] = Adam,
    optimizer_params: Optional[Dict[str, Any]] = None,
    lr_annealing: bool = False,
    warmup_percentage: float = 0,
    verbose: bool = True,
    logger: Optional[Union[Logger, bool]] = None,
    early_stopping: bool = False,
    patience: Optional[int] = None,
    tolerance: float = 0.001,
    fast_train: bool = False,
    train_logged_methods: Optional[
        List[Union[str, Tuple[str, Callable]]]
    ] = None,
    eval_logged_methods: Optional[
        List[Union[str, Tuple[str, Callable]]]
    ] = None,
) -> Optional[Any]:
    # Create the training and validation dataloaders
    train_loader, valid_loader = make_default_dataloaders(
        data=train_data,
        valid_data=valid_data,
        valid_percentage=valid_percentage,
        batch_size=batch_size,
        num_workers=num_workers,
    )

    # Pretrain all the components (such as quantization schemes) that need training
    pretrain_components(
        estimator=estimator, train_loader=train_loader, verbose=verbose
    )

    # Update the dataloader if the estimator requires custom batches
    train_loader = update_dataloader(
        estimator,
        train_loader,
    )

    # And the validation loader (if any)
    if not (valid_loader is None):
        valid_loader = update_dataloader(estimator, valid_loader)

    # Train the model
    return train_model(
        model=estimator,
        train_data=train_loader,
        valid_data=valid_loader,
        valid_percentage=0,
        early_stopping=early_stopping,
        patience=patience,
        tolerance=tolerance,
        batch_size=batch_size,
        max_epochs=max_epochs,
        lr_annealing=lr_annealing,
        warmup_percentage=warmup_percentage,
        max_iterations=max_iterations,
        optimizer_class=optimizer_class,
        optimizer_params=optimizer_params,
        eval_method="mutual_information",
        fast_train=fast_train,
        num_workers=num_workers,
        device=device,
        verbose=verbose,
        logger=logger,
        train_logged_methods=train_logged_methods,
        eval_logged_methods=eval_logged_methods,
    )



from copy import deepcopy
from typing import Callable, TypeVar, Dict, Any
import inspect

import torch
from torch import nn

T = TypeVar("T")


def _is_cache_valid(arguments: Any, stored_arguments: Any):
    if stored_arguments is None:
        return False

    if torch.is_tensor(arguments):
        if not torch.equal(stored_arguments, arguments):
            return False
    elif isinstance(arguments, dict):
        if stored_arguments.keys() != arguments.keys():
            return False
        for k in arguments:
            if not _is_cache_valid(arguments[k], stored_arguments[k]):
                return False
    elif isinstance(arguments, list) or isinstance(arguments, tuple):
        if len(arguments) != len(stored_arguments):
            return False
        for i in range(len(arguments)):
            if not _is_cache_valid(arguments[i], stored_arguments[i]):
                return False
    else:
        if stored_arguments != arguments:
            return False
    return True


def delete_cache_hook(self, *args, **kwargs):
    if hasattr(self, "__cache"):
        del self.__cache


def cached_method(method: Callable[..., T]) -> Callable[..., T]:
    def wrapper(*args, **kwargs):
        arguments = inspect.getcallargs(method, *args, **kwargs)
        self = arguments.pop("self")

        # Attach __cache to the instance
        if not hasattr(self, "__cache"):
            self.__cache = {}
            self.__cache_stats = {}

            if hasattr(self, "__getstate__"):
                original_getstate = self.__getstate__
            else:
                original_getstate = None

            # Wrap the get_state method to exclude the cache
            def wrapped_getstate():
                # delete_cache_hook(self)
                if original_getstate is None:
                    state = self.__dict__
                else:
                    state = original_getstate()

                if "__getstate__" in state:
                    state["__getstate__"] = original_getstate
                if "__cache" in state:
                    del state["__cache"]
                if "__cache_stats" in state:
                    del state["__cache_stats"]
                return state

            self.__getstate__ = wrapped_getstate

            # For nn.Modules, automatically invalidate cache on backwards()
            if isinstance(self, nn.Module):
                self.register_full_backward_hook(delete_cache_hook)

        # Initialize the stats and empty cache
        if not (method.__name__ in self.__cache):
            self.__cache[method.__name__] = None, None

        if not (method.__name__ in self.__cache_stats):
            self.__cache_stats[method.__name__] = {"hits": 0, "miss": 0}

        stored_arguments, stored_value = self.__cache[method.__name__]
        stats = self.__cache_stats[method.__name__]

        if not _is_cache_valid(arguments, stored_arguments):
            value = method(*args, **kwargs)
            self.__cache[method.__name__] = (arguments, value)
            stats["miss"] += 1
        else:
            stats["hits"] += 1
            value = stored_value

        return value

    wrapper.__signature__ = inspect.signature(method)

    return wrapper


def cached_function(method: Callable[..., T]) -> Callable[..., T]:
    stored_arguments = {}
    stored_value = None
    stats = {"hits": 0, "miss": 0}

    def cache_info():
        nonlocal stats
        return stats

    def delete_cache():
        nonlocal stored_value, stored_arguments
        stored_value = None
        stored_arguments = {}

    def wrapper(*args, **kwargs):
        arguments = inspect.getcallargs(method, *args, **kwargs)

        # static cache for functions (no methods)
        nonlocal stats, stored_arguments, stored_value

        if not _is_cache_valid(arguments, stored_arguments):
            value = method(*args, **kwargs)
            stored_value = value
            stored_arguments = arguments
            stats["miss"] += 1
        else:
            stats["hits"] += 1
            value = stored_value

        return value

    wrapper.cache_info = cache_info
    wrapper.delete_cache = delete_cache
    wrapper.__signature__ = inspect.signature(method)
    return wrapper



from __future__ import annotations

import inspect
import os
from abc import abstractmethod
from contextlib import contextmanager
import numpy as np
import torch
from torch import nn
from typing import Any, Callable, Optional, Union, List, Tuple, Dict

from torch_mist.utils.logging.metrics import compute_mean


class Logger:
    def __init__(self, log_dir: str = ".", log_every: int = 1):
        self.log_dir = log_dir
        self.log_every = log_every
        self._buffer = {}
        self._logged_methods = {}
        self._split = None
        self._iteration = 0
        self._epoch = 0

    @contextmanager
    def train(self):
        self.on_split_start("train")
        try:
            yield
        finally:
            self.on_split_end("train")

    @contextmanager
    def valid(self):
        self.on_split_start("valid")
        try:
            yield
        finally:
            self.on_split_end("valid")

    @contextmanager
    def test(self):
        self.on_split_start("test")
        try:
            yield
        finally:
            self.on_split_end("test")

    @contextmanager
    def iteration(self):
        self.on_iteration_start()
        try:
            yield
        finally:
            self.on_iteration_end()

    @contextmanager
    def epoch(self):
        self.on_epoch_start()
        try:
            yield
        finally:
            self.on_epoch_end()

    def on_split_start(self, name: str):
        assert self._split is None
        self._split = name

    def on_split_end(self, name: str):
        if name != "train":
            self._log_buffer(force_logging=name)
        self._split = None

    def on_iteration_start(self):
        pass

    def on_iteration_end(self):
        self._iteration += 1
        self._log_buffer()
        self._log_buffer()

    def on_epoch_start(self):
        self._epoch += 1

    def on_epoch_end(self):
        pass

    @contextmanager
    def logged_methods(
        self,
        instance: Any,
        methods: List[Union[str, Tuple[str, Callable[[Any, Any], Any]]]],
    ):
        for method in methods:
            if isinstance(method, tuple):
                method_name, metric = method
            else:
                method_name = method
                metric = compute_mean
            self.add_logging_hook(instance, method_name, metric)
        try:
            yield
        finally:
            for method in methods:
                if isinstance(method, tuple):
                    method_name, _ = method
                else:
                    method_name = method
                self.detach_hook(method_name)

    def add_logging_hook(
        self,
        instance: Any,
        method_name: str,
        metric: Callable[[Any, Any], Dict[str, Any]],
    ):
        for attr in method_name.split(".")[:-1]:
            instance = getattr(instance, attr)
        full_name = method_name
        method_name = method_name.split(".")[-1]
        original_method = getattr(instance, method_name)
        assert full_name not in self._logged_methods
        self._logged_methods[full_name] = (
            instance,
            method_name,
            original_method,
        )

        def method_with_hook(*args, **kwargs):
            output = original_method(*args, **kwargs)
            arguments = inspect.getcallargs(original_method, *args, **kwargs)

            if self._split:
                stats = metric(input=arguments, output=output)
                if not (stats is None):
                    if not isinstance(stats, dict):
                        stats = {"": stats}
                    self.log(name=full_name, data_dict=stats)

            return output

        method_with_hook.__signature__ = inspect.signature(original_method)

        setattr(instance, method_name, method_with_hook)

    def detach_hook(self, method_name):
        instance, method_original_name, original_method = self._logged_methods[
            method_name
        ]
        setattr(instance, method_original_name, original_method)
        del self._logged_methods[method_name]

    def is_logged(self, method_name: str) -> bool:
        return method_name in self._logged_methods

    def _log_buffer(self, force_logging: Optional[str] = None):
        keys_to_delete = []
        for full_name, (orig_iteration, data) in self._buffer.items():
            split = full_name.split("/")[0]
            name = "/".join(full_name.split("/")[1:])
            if (
                self._iteration - orig_iteration >= self.log_every
                or force_logging == split
            ):
                agg_values = {}
                for metric, values in data.items():
                    first_value = values[0]
                    if (
                        isinstance(first_value, float)
                        or (
                            isinstance(first_value, torch.Tensor)
                            and first_value.numel() == 1
                        )
                        or (
                            isinstance(first_value, np.ndarray)
                            and first_value.size == 1
                        )
                    ):
                        agg_values[metric] = np.mean(values)
                    else:
                        agg_values[metric] = first_value

                # Remove the dictionary wrap if necessary
                if len(agg_values) == 1:
                    if "" in agg_values:
                        agg_values = agg_values[""]

                self._log(
                    data=agg_values,
                    name=name,
                    iteration=self._iteration,
                    epoch=self._epoch,
                    split=split,
                )
                keys_to_delete.append(full_name)

        for full_name in keys_to_delete:
            del self._buffer[full_name]

    def log(self, name: str, data_dict: Any):
        name = f"{self._split}/{name}"
        if not (name in self._buffer):
            self._buffer[name] = (
                self._iteration,
                {metric: [value] for metric, value in data_dict.items()},
            )
        else:
            for metric, value in data_dict.items():
                self._buffer[name][1][metric].append(value)

    @abstractmethod
    def _log(
        self, data: Any, name: str, iteration: int, epoch: int, split: str
    ):
        raise NotImplementedError()

    @abstractmethod
    def _reset_log(self):
        raise NotImplementedError()

    def reset_log(self):
        self._reset_log()
        self._split = None
        self._iteration = 0
        self._epoch = 0

    def detach_all_hooks(self):
        logged_methods = list(self._logged_methods)
        for method_name in logged_methods:
            self.detach_hook(method_name)

    def clear(self):
        self.detach_all_hooks()
        self.reset_log()

    def __del__(self):
        self.clear()

    def get_log(self) -> Optional[Any]:
        return None

    def _save_log(self, log: Any):
        pass

    def save_log(self):
        log = self.get_log()
        self._save_log(log)

    @abstractmethod
    def save_model(self, model: nn.Module, name: str):
        filepath = os.path.join(self.log_dir, name)
        torch.save(model, filepath)


class DummyLogger(Logger):
    def _log(
        self, data: Any, name: str, iteration: int, epoch: int, split: str
    ):
        pass

    def _reset_log(self):
        pass



from typing import Optional, Any, Dict

import torch
from pyro.distributions import ConditionalDistribution
from torch import nn

from torch_mist.nn import Model
from torch_mist.quantization import QuantizationFunction
from torch_mist.quantization.functions import (
    LearnableVectorQuantization,
)

VERSIONS = ["v1", "v2"]


class VQVAE(LearnableVectorQuantization, Model):
    upper_bound: bool = True

    def __init__(
        self,
        encoder: nn.Module,
        decoder: ConditionalDistribution,
        initial_vectors: Optional[torch.Tensor] = None,
        n_bins: Optional[int] = None,
        quantization_dim: Optional[int] = None,
        beta: float = 0.25,
        gamma: float = 0.99,
        version: str = "v2",
        **train_params,
    ):
        if not (version in VERSIONS):
            raise ValueError(f"Please choose version among {VERSIONS}.")
        LearnableVectorQuantization.__init__(
            self,
            vectors=initial_vectors,
            n_bins=n_bins,
            quantization_dim=quantization_dim,
            **train_params,
        )
        self.encoder = encoder
        self.decoder = decoder

        self.beta = beta
        self.gamma = gamma
        self.version = version

        initial_assignments = torch.ones(self.n_bins) / self.n_bins
        self.register_buffer("assignments", initial_assignments)
        self.min_assignment = 0.1 / self.n_bins

    def quantize(self, x: torch.Tensor) -> torch.LongTensor:
        z = self.encoder(x)
        assert z.shape[-1] == self.vectors.shape[-1]
        return super().quantize(z)

    def _update_vectors(self, indices: torch.LongTensor, z: torch.Tensor):
        # Compute the one-hot representation of the indices [N*M, n_bins]
        flat_indices = indices.view(-1)
        one_hot = torch.nn.functional.one_hot(
            flat_indices, self.n_bins
        ).float()

        # Compute the number of datapoints assigned to each bin [n_bins]
        assignments = one_hot.mean(0)

        self.assignments = (
            self.gamma * self.assignments + (1 - self.gamma) * assignments
        )

        # Compute the mean vector for each assignment [n_bins, z_dim]
        N = flat_indices.shape[0]
        mean_vector = (one_hot.unsqueeze(-1) * z.unsqueeze(-2)).sum(0)
        mean_vector = mean_vector / (assignments.unsqueeze(-1) * N)

        not_assigned = [
            i.item()
            for i in torch.arange(self.n_bins)[
                self.assignments < self.min_assignment
            ].long()
        ]

        while len(not_assigned) > 0:
            not_assigned_idx = not_assigned.pop()
            # Select the vector with the highest assignment
            max_idx = torch.argmax(self.assignments)
            # set the centroid to the vector with the highest assignment + some noise
            noise = torch.randn_like(self.vectors[max_idx]) * 1e-5
            self.vectors[not_assigned_idx] = self.vectors[max_idx] + noise
            # set the new assignements to half the max assignment
            new_assignment = self.assignments[max_idx] / 2
            self.assignments[not_assigned_idx] = new_assignment
            self.assignments[max_idx] = new_assignment

        mean_vector[assignments == 0] = self.vectors[assignments == 0]

        self.vectors = (
            self.gamma * self.vectors + (1 - self.gamma) * mean_vector
        )

    def loss(self, x: torch.Tensor) -> torch.Tensor:
        z = self.encoder(x)
        indices = super().quantize(z)

        vectors = self.vectors[indices]
        z_q = z + (vectors - z).detach()

        assert (
            z_q.shape == z.shape == vectors.shape
        ), f"{z_q.shape} != {z.shape} != {vectors.shape}"

        reconstruction_loss = -self.decoder.condition(z_q).log_prob(x)
        assert (
            reconstruction_loss.shape == x.shape[:-1]
        ), f"{reconstruction_loss.shape} != {x.shape[:-1]}"

        reconstruction_loss = reconstruction_loss.mean()
        commitment_loss = torch.mean((vectors.detach() - z) ** 2)

        loss = reconstruction_loss + self.beta * commitment_loss

        if self.training and not self.trained:
            if self.version == "v1":
                codebook_loss = torch.mean((vectors - z.detach()) ** 2)
                loss += codebook_loss
            elif self.version == "v2":
                self._update_vectors(indices, z)
            else:
                raise NotImplementedError()

        return loss



from abc import abstractmethod

import torch
from pyro.distributions import ConditionalDistribution

from torch_mist.distributions import JointDistribution
from torch_mist.estimators.base import MIEstimator
from torch_mist.utils.caching import cached_method
from torch_mist.utils.shape import expand_to_same_shape


class GenerativeMIEstimator(MIEstimator):
    def unnormalized_log_ratio(
        self, x: torch.Tensor, y: torch.Tensor
    ) -> torch.Tensor:
        return self.log_ratio(x, y)

    @abstractmethod
    def approx_log_p_y_given_x(
        self, x: torch.Tensor, y: torch.Tensor
    ) -> torch.Tensor:
        raise NotImplementedError()

    @abstractmethod
    def approx_log_p_y(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError()

    def log_ratio(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
    ) -> torch.Tensor:
        x_dim = x.ndim + (1 if not torch.is_floating_point(x) else 0)
        y_dim = y.ndim + (1 if not torch.is_floating_point(y) else 0)
        assert x_dim == y_dim, f"x.ndim={x_dim}, y.ndim={y_dim}"

        # Compute the ratio using the primal KL bound
        approx_log_p_y_given_x = self.approx_log_p_y_given_x(x=x, y=y)
        approx_log_p_y = self.approx_log_p_y(x=x, y=y)

        assert (
            approx_log_p_y_given_x.ndim == approx_log_p_y.ndim == x_dim - 1
        ), f"log_p_y_x.ndim={approx_log_p_y_given_x.ndim}, log_p_y.ndim={approx_log_p_y.ndim}"
        log_ratio = approx_log_p_y_given_x - approx_log_p_y

        return log_ratio


class ConditionalGenerativeMIEstimator(GenerativeMIEstimator):
    def __init__(self, q_Y_given_X: ConditionalDistribution):
        super().__init__()
        # Add caching to the conditioning for efficiency
        assert isinstance(q_Y_given_X, ConditionalDistribution)
        self.q_Y_given_X = q_Y_given_X

    @cached_method
    def approx_log_p_y_given_x(
        self, x: torch.Tensor, y: torch.Tensor
    ) -> torch.Tensor:
        # Expand to the same shape
        x, y = expand_to_same_shape(x, y)

        q_Y_given_x = self.q_Y_given_X.condition(x)

        # Compute log q(Y=y|X=x)]
        log_q_y_given_x = q_Y_given_x.log_prob(y)

        assert (
            log_q_y_given_x.shape == y.shape[:-1]
            and torch.is_floating_point(y)
        ) or (
            log_q_y_given_x.shape == y.shape and not torch.is_floating_point(y)
        ), f"log_p_Y_X.shape={log_q_y_given_x.shape}, y.shape={y.shape}"

        return log_q_y_given_x

    def batch_loss(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        loss = -self.approx_log_p_y_given_x(x=x, y=y)

        assert (loss.shape == y.shape[:-1] and torch.is_floating_point(y)) or (
            loss.shape == y.shape and not torch.is_floating_point(y)
        ), f"{torch.is_floating_point(y)}. {loss.shape}!={y.shape[:-1]}"
        return loss

    def __repr__(self):
        s = self.__class__.__name__ + "(\n"
        s += (
            "  "
            + "(q_Y_given_X): "
            + str(self.q_Y_given_X).replace("\n", "  \n")
            + "\n"
        )
        s += ")" + "\n"

        return s


class JointGenerativeMIEstimator(GenerativeMIEstimator):
    def __init__(self, q_XY: JointDistribution):
        super().__init__()
        # Add caching to the conditioning for efficiency
        assert isinstance(q_XY, JointDistribution)
        self.q_XY = q_XY

    @property
    def q_X(self):
        return self.q_XY.marginal("x")

    @property
    def q_Y(self):
        return self.q_XY.marginal("y")

    @cached_method
    def approx_log_p_xy(
        self, x: torch.Tensor, y: torch.Tensor
    ) -> torch.Tensor:
        # The shape is [...]
        log_q_xy = self.q_XY.log_prob(x=x, y=y)

        assert (
            log_q_xy.shape == y.shape[:-1] and torch.is_floating_point(y)
        ) or (log_q_xy.shape == y.shape and not torch.is_floating_point(y))

        return log_q_xy

    @cached_method
    def approx_log_p_x(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        log_q_x = self.q_X.log_prob(x)
        assert (
            log_q_x.shape == x.shape[:-1] and torch.is_floating_point(x)
        ) or (log_q_x.shape == x.shape and not torch.is_floating_point(x))
        # The shape is [...]
        return log_q_x

    @cached_method
    def approx_log_p_y(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        log_q_y = self.q_Y.log_prob(y)
        assert (
            log_q_y.shape == y.shape[:-1] and torch.is_floating_point(y)
        ) or (log_q_y.shape == y.shape and not torch.is_floating_point(y))
        # The shape is [...]
        return log_q_y

    @cached_method
    def approx_log_p_y_given_x(
        self, x: torch.Tensor, y: torch.Tensor
    ) -> torch.Tensor:
        log_q_xy = self.approx_log_p_xy(x, y)
        log_q_x = self.approx_log_p_x(x, y)

        # Compute log q(Y=y|X=x)]
        log_q_y_given_x = log_q_xy - log_q_x

        assert (
            log_q_y_given_x.shape == y.shape[:-1]
            and torch.is_floating_point(y)
        ) or (
            log_q_y_given_x.shape == y.shape and not torch.is_floating_point(y)
        ), f"log_p_Y_X.shape={log_q_y_given_x.shape}, y.shape={y.shape}"

        return log_q_y_given_x

    def batch_loss(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        loss = -self.approx_log_p_xy(x=x, y=y)
        assert (loss.shape == y.shape[:-1] and torch.is_floating_point(y)) or (
            loss.shape == y.shape and not torch.is_floating_point(y)
        )
        return loss

    def __repr__(self):
        s = self.__class__.__name__ + "(\n"
        s += "  (q_XY): " + str(self.q_XY).replace("\n", "\n  ") + "\n"
        if self.q_X is not None:
            s += "  (q_X): " + str(self.q_X).replace("\n", "\n  ") + "\n"
        if self.q_Y is not None:
            s += "  (q_Y): " + str(self.q_Y).replace("\n", "\n  ") + "\n"
        s += ")" + "\n"
        return s



from typing import List, Tuple, Dict, Callable

import torch
from torch import nn

from torch_mist.estimators.base import MIEstimator


class MultiMIEstimator(MIEstimator):
    def __init__(self, estimators: Dict[Tuple[str, str], MIEstimator]):
        super().__init__()

        self.estimators = nn.ModuleDict()
        self.infomax_gradient = {}
        self.lower_bound = True
        self.upper_bound = True
        for (x_key, y_key), estimator in estimators.items():
            self.upper_bound = self.upper_bound and estimator.upper_bound
            self.lower_bound = self.lower_bound and estimator.lower_bound

            # Add the renaming for the components that need pre-training (if needed)
            for data_process, component in estimator._components_to_pretrain:

                def new_data_process(batch: Dict[str, torch.Tensor]):
                    batch["x"] = batch[x_key]
                    batch["y"] = batch[y_key]
                    return data_process(batch)

                self._components_to_pretrain.append(
                    (new_data_process, component)
                )

            key = f"{x_key};{y_key}"
            self.estimators[key] = estimator
            if x_key in self.infomax_gradient:
                self.infomax_gradient[x_key] = (
                    self.infomax_gradient[x_key]
                    and estimator.infomax_gradient["x"]
                )
            else:
                self.infomax_gradient[x_key] = estimator.infomax_gradient["x"]
            if y_key in self.infomax_gradient:
                self.infomax_gradient[y_key] = (
                    self.infomax_gradient[y_key]
                    and estimator.infomax_gradient["y"]
                )
            else:
                self.infomax_gradient[y_key] = estimator.infomax_gradient["y"]

    def broadcast_function(
        self, function_name: str, **variables
    ) -> Dict[Tuple[str, str], torch.Tensor]:
        values = {}
        for key, estimator in self.estimators.items():
            x_key, y_key = key.split(";")
            if not (x_key in variables):
                raise ValueError(f"The variable {x_key} is not specified.")
            if not (y_key in variables):
                raise ValueError(f"The variable {y_key} is not specified.")

            x = variables[x_key]
            y = variables[y_key]

            value = getattr(estimator, function_name)(x, y)
            values[(x_key, y_key)] = value
        return values

    def loss(self, **variables) -> torch.Tensor:
        losses = self.broadcast_function("loss", **variables)
        total_loss: torch.Tensor = 0
        for loss in losses.values():
            total_loss += loss
        return total_loss

    def batch_loss(self, **variables) -> torch.Tensor:
        losses = self.broadcast_function("batch_loss", **variables)
        total_loss: torch.Tensor = 0
        for loss in losses.values():
            total_loss += loss
        return total_loss

    def mutual_information(
        self, **variables
    ) -> Dict[Tuple[str, str], torch.Tensor]:
        return self.broadcast_function("mutual_information", **variables)

    def log_ratio(self, **variables) -> Dict[Tuple[str, str], torch.Tensor]:
        return self.broadcast_function("log_ratio", **variables)

    def forward(self, **variables) -> torch.Tensor:
        return self.loss(**variables)



from typing import Union, List

import torch

from torch_mist.estimators import TransformedMIEstimator, MIEstimator


class FlippedMIEstimator(TransformedMIEstimator):
    def __init__(self, base_estimator: MIEstimator):
        super().__init__(base_estimator=base_estimator)

    def log_ratio(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return self.base_estimator.log_ratio(x=y, y=x)

    def unnormalized_log_ratio(
        self, x: torch.Tensor, y: torch.Tensor
    ) -> torch.Tensor:
        return self.base_estimator.unnormalized_log_ratio(x=y, y=x)

    def batch_loss(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return self.base_estimator.batch_loss(x=y, y=x)

    def loss(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return self.base_estimator.loss(x=y, y=x)

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return self.base_estimator.forward(x=y, y=x)



import pyro.distributions.transforms as transforms
import torch
from pyro.distributions import TransformModule

from torch import nn


class Permute(transforms.Permute, TransformModule):
    def __init__(self, permutation, *, dim=-1, cache_size=1):
        nn.Module.__init__(self)
        transforms.Permute.__init__(
            self, permutation=permutation, dim=dim, cache_size=cache_size
        )
        self.register_buffer("placeholder", torch.zeros(1))

    def update_device(self):
        if self.permutation.device != self.placeholder.device:
            self.permutation = self.permutation.to(self.placeholder.device)

    def _call(self, x):
        self.update_device()
        return super(Permute, self)._call(x)

    def _inverse(self, y):
        self.update_device()
        return super(Permute, self)._inverse(y)

    def log_abs_det_jacobian(self, x, y):
        self.update_device()
        return super(Permute, self).log_abs_det_jacobian(x, y)



from abc import abstractmethod
from typing import Optional, Tuple

import torch

from torch_mist.estimators.base import MIEstimator
from torch_mist.estimators.discriminative.base import DiscriminativeMIEstimator
from torch_mist.utils.caching import cached_method
from torch_mist.utils.freeze import is_trainable
from contextlib import contextmanager


class HybridMIEstimator(DiscriminativeMIEstimator):
    def __init__(
        self,
        generative_estimator: MIEstimator,
        discriminative_estimator: DiscriminativeMIEstimator,
    ):
        neg_samples = discriminative_estimator.neg_samples
        super().__init__(
            critic=discriminative_estimator.critic,
            neg_samples=neg_samples,
        )

        self.discriminative_estimator = discriminative_estimator
        self.generative_estimator = generative_estimator

        informax_gradient = generative_estimator.infomax_gradient
        informax_gradient = {
            key: value and discriminative_estimator.infomax_gradient[key]
            for key, value in informax_gradient.items()
        }
        self.infomax_gradient = informax_gradient

        self._components_to_pretrain += (
            self.generative_estimator._components_to_pretrain
        )
        self._components_to_pretrain += (
            self.discriminative_estimator._components_to_pretrain
        )

    @cached_method
    def unnormalized_log_ratio(
        self, x: torch.Tensor, y: torch.Tensor
    ) -> torch.Tensor:
        f = self.discriminative_estimator.unnormalized_log_ratio(x, y)
        partial_log_ratio = self.generative_estimator.log_ratio(x, y)

        assert f.shape == partial_log_ratio.shape
        return f + partial_log_ratio

    @cached_method
    def log_ratio(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        with self.resampling_strategy():
            log_rest = self.discriminative_estimator.log_ratio(x, y)
        partial_log_ratio = self.generative_estimator.log_ratio(x, y)

        assert partial_log_ratio.shape == log_rest.shape
        return partial_log_ratio + log_rest

    @abstractmethod
    def sample_negatives(
        self, x: torch.Tensor, y: torch.Tensor
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        raise NotImplementedError()

    @contextmanager
    def resampling_strategy(self):
        # Store the original sampling scheme
        original_method = self.discriminative_estimator.sample_negatives

        # Replace it with the new one
        self.discriminative_estimator.sample_negatives = self.sample_negatives
        try:
            yield
        finally:
            # Restore the original method
            self.discriminative_estimator.sample_negatives = original_method

    def batch_loss(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        # Compute only the discriminative loss component
        # Note that we can skip computing the log r(y|x)/p(y)
        with self.resampling_strategy():
            batch_loss = self.discriminative_estimator.batch_loss(x, y)

        # If the generative component is not trainable, there is no need to compute the log-ratio or the generative loss
        if is_trainable(self.generative_estimator):
            batch_loss += self.generative_estimator.batch_loss(x, y)

        return batch_loss

    def __repr__(self):
        s = self.__class__.__name__ + "(\n"
        s += (
            "  (generative_estimator): "
            + str(self.generative_estimator).replace("\n", "\n" + "  ")
            + "\n"
        )
        s += (
            "  (discriminative_estimator): "
            + str(self.discriminative_estimator).replace("\n", "\n" + "  ")
            + "\n"
        )
        s += ")"

        return s
